% $Id: fs2desc.pl,v 1.3 2002/10/14 20:31:52 dm Exp $
% 
% Defines a pretty printer which prints TRALE descriptions.
% Author: Gerald Penn 
%
% $Log: fs2desc.pl,v $
% Revision 1.3  2002/10/14 20:31:52  dm
% extensionalise/3 is now extensionalise/2 - changed call to it here
%
% Revision 1.2  2002/08/07 04:42:58  dm
% Added fs2desc/2 predicate as top-level goal here.
%

:- use_module(library(assoc)).
:- use_module(library(terms),[subsumes_chk/2]).

% ================================================
% fs2desc(+FS:<fs>, -Desc:<desc>) 
%
% Returns the ALE description Desc which is the
% most general satisfier of FS with inequalities Iqs
%
% Attention: This destroys the FS used as input since
% fully_deref overwrites the Tags in its input.

fs2desc(Ref0-SVs0,Desc) :-
   extensionalise(Ref0,SVs0),
   fully_deref(Ref0,SVs0,Ref,SVs),
   pp_desc(Ref,SVs,Desc).

% ------------------------------------------------------------------------------
% pp_desc(+Tag,+SVs,-Desc)
% pp_desc(+Tag,+SVs,+Iqs,-Desc,+PredName,-IqGoal,-IqPreds)
% ------------------------------------------------------------------------------
% Convert the feature structure Tag-SVs/Iqs to a description, Desc, logic
% program IqPreds, and goal, IqGoal, such that query((X=Desc,IqGoal))
% over IqPreds iteratively succeeds with X bound to an alpha-variant of
% Tag-SVs, non-deterministically enumerating every satisfying selection of
% conjuncts from Iqs.
%   In the absence of extensional types, there is a unique selection, in which
% case X is isomorphic to Tag-SVs/Iqs in the inequational algebra.  In the
% absence of inequations, the most general satisfier of Desc is Tag-SVs.  Desc
% is the minimal description, relative to the type system cum appropriateness,
% whose most general satisfier is Tag-SVs, although there may exist smaller
% descriptions in the presence of universal constraints.
%
% Tag-SVs/Iqs is assumed to have been extensionalised, fully dereferenced,
% and pruned.  This can be achieved by calling:
%   extensionalise(Tag0,SVs0),
%   fully_deref_prune(Tag0,SVs0,Tag,SVs,Iqs0,Iqs)
% before calling pp_desc(Tag,SVs,Iqs,_,_,_,_).
%
% The caller must choose the name of the logic programming predicate, PredName,
% to accompany the description.
% ------------------------------------------------------------------------------

pp_desc(Tag,SVs,Desc) :-
  pp_desc(Tag,SVs,[],Desc,_,_,_).

pp_desc(Tag,SVs,Iqs,Desc,PredName,IqGoal,IqPreds) :-
  empty_assoc(VarsIn),
  ppd_dups(Tag,SVs,Iqs,VarsIn,VarsMid2),
% DESCRIPTION GENERATION
  ppd_fs(Tag,SVs,bot,VarsMid2,VarsOut,Desc,desc),
  (Iqs == [] -> IqGoal = true, IqPreds = []
% INEQUATION GENERATION
% this is where we also add a two-place inequation relation - probably should
% be done elsewhere.
  ; IqPreds = [(IqHead :- IqBody),(ineq(X,(=\= X)) :- true)],
    ppd_iqs(Iqs,VarsOut,IqBody0),
    term_variables(IqBody0,IqBVars),
% the purpose of IqGoal is to pass the variables shared by Desc and the
% inequations into the binding scope of IqPreds.
    desc_bound_vars(IqBVars,VarsOut,GoalArgs),
% so far, we have been using description variables in the body.  Ideally,
% IqPreds should not share structure at compile-time with Desc and IqGoal,
% however, in case one or the other's variables are instantiated.
    copy_term([GoalArgs|IqBody0],[HeadArgs|IqBody]),
    IqGoal =.. [PredName|GoalArgs],
    IqHead =.. [PredName|HeadArgs]
  ).

% ------------------------------------------------------------------------------
% ppd_dups(+Tag,+SVs,+Iqs,+VarsIn,-VarsOut)
% ------------------------------------------------------------------------------
% Traverse Tag-SVs/Iqs to create an association list of reference tags.  If
% a tag appears more than once, its value in VarsOut is bound to a pair
% [DVar|Visited], where DVar is a description variable, and Visited is a
% flag used to indicate whether a tag is visited first by the description
% generator or the inequation generator.
% ------------------------------------------------------------------------------
ppd_dups(Tag,SVs,Iqs,VarsIn,VarsOut) :-
  ppd_dups_fs(Tag,SVs,VarsIn,VarsMid),
  ppd_dups_iqs(Iqs,VarsMid,VarsOut).

ppd_dups_fs(Tag,SVs,VarsIn,VarsOut) :-
  get_assoc(Tag,VarsIn,var(_,_,_)) -> VarsOut = VarsIn
  ; put_assoc(Tag,VarsIn,_,VarsMid),
    (SVs = (a_ _) -> VarsOut = VarsMid
    ; SVs =.. [_|Vs],
      ppd_dups_vs(Vs,VarsMid,VarsOut)
    ).

ppd_dups_vs([],Vars,Vars).
ppd_dups_vs([Tag-SVs|Vs],VarsIn,VarsOut) :-
  ppd_dups_fs(Tag,SVs,VarsIn,VarsMid),
  ppd_dups_vs(Vs,VarsMid,VarsOut).

ppd_dups_iqs([],Vars,Vars).
ppd_dups_iqs([Iq|Iqs],VarsIn,VarsOut) :-
  ppd_dups_ineq(Iq,VarsIn,VarsMid),
  ppd_dups_ineq(Iqs,VarsMid,VarsOut).

ppd_dups_ineq(done,Vars,Vars).
ppd_dups_ineq(ineq(Tag1,SVs1,Tag2,SVs2,Ineqs),VarsIn,VarsOut) :-
  ppd_dups_fs(Tag1,SVs1,VarsIn,VarsMid),
  ppd_dups_fs(Tag2,SVs2,VarsMid,VarsMid2),
  ppd_dups_ineq(Ineqs,VarsMid2,VarsOut).

% ------------------------------------------------------------------------------
% ppd_fs(+Tag,+SVs,+R,+VarsIn,-VarsOut,-Desc,-VisFlag)
% ------------------------------------------------------------------------------
% Desc is a minimal description whose most general satisfier is Tag-SVs,
% in the context of type inferencing mode, R.  VisFlag indicates whether
% this predicate was called in the context of description or inequation
% generation.  VarsOut is VarsIn, with the type modes of variables in Desc
% updated.
% ------------------------------------------------------------------------------

ppd_fs(Tag,SVs,R,VarsIn,VarsOut,Desc,VisFlag) :-
  get_assoc(Tag,VarsIn,TagBind),
% if this is a singular tag, don't bind a variable to it
  ( TagBind = nobind -> ppd_svs(SVs,R,VarsIn,VarsOut,Descs,VisFlag),
                        list_to_desc(Descs,Desc)
  ; TagBind = var(DVar,VMode,Visited),
% otherwise, if this is our first visit, mark with VisFlag, and bind DVar
    ( var(VMode) -> VMode = bot ; true),  % initialise type mode of DVar
    unify_type(R,VMode,RNew),  % unify with current type mode
    ( sub_type(RNew,VMode) -> VarsMid = VarsIn  
    ; put_assoc(Tag,VarsIn,var(DVar,RNew,Visited),VarsMid) 
    ),  % update type mode of DVar
    ( var(Visited) -> Visited = VisFlag,
	              ppd_svs(SVs,RNew,VarsMid,VarsOut,Descs,VisFlag),
                      list_to_desc(Descs,DVar,Desc)
% if it's not our first visit, use DVar only - the first visit is 
%  responsible for the rest of the description.
    ; Desc = DVar, VarsOut = VarsMid
    )
  ).

ppd_svs(SVs,R,VarsIn,VarsOut,Descs,VisFlag) :-
  (SVs = (a_ _) -> Type = SVs, Vs = [], FDescs = [], TypInf = R,
                   VarsOut = VarsIn
  ; SVs =.. [Type|Vs],
    approps(Type,FRs,_),
    ppd_vs(Vs,FRs,R,TypInf,VarsIn,VarsOut,FDescs,VisFlag)
% if the type can be inferred from the mode plus type inferencing on
% feature names, leave the type out.
  ),
  (sub_type(Type,TypInf) -> Descs = FDescs ; Descs = [Type|FDescs]).

ppd_vs([],[],TypInf,TypInf,Vars,Vars,[],_).
ppd_vs([Tag-SVs|Vs],[F:R|FRs],TypIn,TypOut,VarsIn,VarsOut,FDescs,VisFlag) :-
  ppd_fs(Tag,SVs,R,VarsIn,VarsMid,FDesc,VisFlag),
% bot means that it is not necessary to add information to this feature
% in this context.
  ( FDesc == bot -> ppd_vs(Vs,FRs,TypIn,TypOut,VarsMid,VarsOut,FDescs,VisFlag)
  ; introduce(F,FIntro),
    unify_type(FIntro,TypIn,TypMid),
    FDescs = [F:FDesc|FDescsRest],
    ppd_vs(Vs,FRs,TypMid,TypOut,VarsMid,VarsOut,FDescsRest,VisFlag)
  ).

% ------------------------------------------------------------------------------
% ppd_iqs(+Iqs,+Vars,-IqBody)
% ------------------------------------------------------------------------------
% Convert a non-empty set of inequations into a CNF of ineq/2 calls that
% will enumerate the acceptable conjunctions of inequations.  This is the
% best we can do in general, because a disjunction may arise from decomposition
% of an inequation attached to an extensionally typed FS not accessible from
% Tag-SVs, and there is no two-place inequation relation in the description
% language.  We could be a little more careful about variable type mode here,
% but that would involve generalising the types of variables after every
% disjunction.
% ------------------------------------------------------------------------------

ppd_iqs([Iq|Iqs],Vars,IqBody) :-
  ppd_iqs(Iqs,Iq,Vars,IqBody).

ppd_iqs([],Iq,Vars,IqBody) :-
  ppd_iq(Iq,Vars,IqBody).
ppd_iqs([Iq|Iqs],Iq0,Vars,(IqBody0,IqBody)) :-
  ppd_iq(Iq0,Vars,IqBody0),
  ppd_iqs(Iqs,Iq,Vars,IqBody).

ppd_iq(ineq(Tag1,SVs1,Tag2,SVs2,Ineqs),Vars,IqBody) :-
  ppd_iq(Ineqs,Tag1,SVs1,Tag2,SVs2,Vars,IqBody).

ppd_iq(done,Tag1,SVs1,Tag2,SVs2,Vars,ineq(Desc1,Desc2)) :-
  ppd_fs(Tag1,SVs1,bot,Vars,VarsMid,Desc1,ineq),
  ppd_fs(Tag2,SVs2,bot,VarsMid,_,Desc2,ineq).
ppd_iq(ineq(Tag3,SVs3,Tag4,SVs4,Ineqs),Tag1,SVs1,Tag2,SVs2,Vars,
       (ineq(Desc1,Desc2);IqBody)) :-
  ppd_fs(Tag1,SVs1,bot,Vars,VarsMid,Desc1,ineq),
  ppd_fs(Tag2,SVs2,bot,VarsMid,_,Desc2,ineq),
  ppd_iq(Ineqs,Tag3,SVs3,Tag4,SVs4,Vars,IqBody).

% ==============================================================================
% UTILITIES
% ==============================================================================

% convert list to conjunctive ALE description
list_to_desc([],bot).     
list_to_desc([Desc|Descs],D0) :-
  list_to_desc(Descs,Desc,D0).

list_to_desc([],Desc,Desc).
list_to_desc([Desc|Descs],D0,(D0,DRest)) :-
  list_to_desc(Descs,Desc,DRest).  

% isolate those Vs that were first visited by the description generator.
desc_bound_vars([],_,[]).
desc_bound_vars([V|Vs],Vars,DBs) :-
  get_assoc(V,Vars,var(_,_,Visited)),
  (Visited == desc -> DBs = [V|DBsRest] ; DBsRest = DBs),
  desc_bound_vars(Vs,Vars,DBsRest).
