%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%   $RCSfile: grammar_interface.pl,v $
%%  $Revision: 1.9 $
%%      $Date: 2003/05/20 20:40:08 $
%%     Author: Detmar Meurers (dm@ling.osu.edu) with additions 
%%             by Stefan Mueller (stefan.mueller@uni-bremen.de)
%%    Purpose: Code to provide a user definable grammar interface
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
%
% Define a way for user to specify what features are used 
% in a given grammar, e.g. for encoding the phonology and 
% the daughters of a phrase - and provide the defaults phon
% and dtrs if nothing is defined.
%
% -----------------------------------
%
% The following predicates can be specified in a grammar:
% 
% 1) For the feature reflecting the properties encoded in the
%    phrase structure (i.e. extra-logically):
% 
%    dtrs_feat(<featurename>).
%    phon_feat(<featurename>).
%
% 2) For finding the semantics (relevant for comparing analyses): 
%
%   cont_path(<path>). (where <path> is a Prolog list of <featurename>)
%   event_type(<typename>).
%   index_type(<typename>).
%
% 3) In principle the user can also define the following
%    list types and features, but the trale code largely
%    has those hardwired so it's not advisable to change 
%    it until we get around to changing all old code
%    to take these predicates into account.
%
% list_type(<typename>).
% ne_list_type(<typename>).
% hd_feat(<featurename>).
% tl_feat(<featurename>).
%
% -----------------------------------
%
% The internal predicates for querying these user definitions
% have the same predicate name, plus a trailing _, i.e.
% the dtrs feature can be obtained by calling dtrs_feat_/1.
% A default is provided for all these predicates, so that
% the return a value regardless of whether the user specified
% them or not.
% ------------------------------------------------------

dtrs_feat_(Feat) :-
   ( current_predicate(dtrs_feat,dtrs_feat(_)) ->
      dtrs_feat(Feat)
   ;   %% The default if the user hasn't defined dtrs_feat/1
      Feat = dtrs
   ).

phon_feat_(Feat) :-
   ( current_predicate(phon_feat,phon_feat(_)) ->
      phon_feat(Feat)
   ;   %% The default if the user hasn't defined phon_feat/1
      Feat = phon
   ).

% ------------------------------------------------------

list_type_(Type) :-
   ( current_predicate(list_type,list_type(_)) ->
      list_type(Type)
   ;   %% The default if the user hasn't defined list_type/1
      Type = list
   ).

ne_list_type_(Type) :-
   ( current_predicate(ne_list_type,ne_list_type(_)) ->
      ne_list_type(Type)
   ;   %% The default if the user hasn't defined ne_list_type/1
      Type = ne_list
   ).

e_list_type_(Type) :-
   ( current_predicate(e_list_type,e_list_type(_)) ->
      e_list_type(Type)
   ;   %% The default if the user hasn't defined e_list_type/1
      Type = e_list
   ).

hd_feat_(Feat) :-
   ( current_predicate(hd_feat,hd_feat(_)) ->
      hd_feat(Feat)
   ;   %% The default if the user hasn't defined hd_feat/1
      Feat = hd
   ).

tl_feat_(Feat) :-
   ( current_predicate(tl_feat,tl_feat(_)) ->
      tl_feat(Feat)
   ;   %% The default if the user hasn't defined tl_feat/1
      Feat = tl
   ).

cont_path_(Path) :-
   ( current_predicate(cont_path,cont_path(_)) ->
      cont_path(Path)
   ;   %% The default if the user hasn't defined cont_path/1
      Path = [synsem,loc,cont]
   ).

event_type_(Type) :-
   ( current_predicate(event_type,event_type(_)) ->
      event_type(Type)
   ;   %% The default if the user hasn't defined event_type/1
      Type = event
   ).

index_type_(Type) :-
   ( current_predicate(index_type,index_type(_)) ->
      index_type(Type)
   ;   %% The default if the user hasn't defined index_type/1
      Type = index
   ).

pred_feat_(Feat) :-
   ( current_predicate(pred_feat,pred_feat(_)) ->
      pred_feat(Feat)
   ;   %% The default if the user hasn't defined pred_feat/1
      Feat = pred
   ).

ind_feat_(Feat) :-
   ( current_predicate(ind_feat,ind_feat(_)) ->
      ind_feat(Feat)
   ;   %% The default if the user hasn't defined pred_feat/1
      Feat = ind
   ).


%%% ==================================================================
% Local Variables:
% mode: prolog
% End:

