% $Id: step_four.pl,v 1.10 2002/12/07 18:34:54 dm Exp $

:- dynamic lex_entry/1.
:- dynamic canon_rep/2.
:- dynamic final_state/1.

%%%In step four, the automata created in the last step are translated
%%%into interaction predicates, if an identical interaction predicate has
%%%not already been created.  All lexical entries with identical automata
%%%will call the same interaction predicate, thus automatically grouping
%%%the lexical entries into natural classes.  Specifically, lexical
%%%items are licensed by a functional relation.

%%%foo ---> q1_0((a:b,
%%%                  b:minus,
%%%                  c:(t2,
%%%                     w:minus,
%%%                     x:minus,
%%%                     y:minus,
%%%                     z:(hd:a,
%%%                        tl:(hd:b,
%%%                            tl:e_list))))).

%%%Representation of interaction predicates:
%%% qClassID(State,InputLexEntry,DerivedLexEntry),
%%% where every interaction predicate has a different functor name,
%%%derived by concatenating 'q' to the identifier of the lexical fsa for
%%%that class.

%%%The lexical entry associated with a state (including the start
%%%state) is licensed by having every state be a final state.
%%%Or, lexical rule application is licensed this way:

%%%fun q1_0(+,-).
%%%q1_0(A, A) if true.
%%%q1_0(A, B) if
%%%        lr_one1(A, C),
%%%        q1_1(C, B).

%%%This new lexicon is written to a file lr_output.pl and re-compiled.
step_four :-
   init_step_four,
   step_four_aux.

init_step_four :-
   retractall(q(_,_,_,_)),
   retractall(canonrep(_,_)),
   retractall(lex_entry(_)).

step_four_aux :-
   lex_info(Phon,LexID,Lex),
   make_new_lexent(Phon,LexID,Lex),
   fail
   ;
   true.

%%%make_new_lexent,2 takes the phon value and the lexical id, gets the
%%%class id for that lexical entry (not caring whether a new interaction
%%%predicate has been created or an old one is being used), and creates
%%%the new lexical entry, where a phon value calls the interaction
%%%predicate for its class.

make_new_lexent(Phon,LexID,Lex) :-
   lex_class(LexID,PredID0),
   concat(PredID0,'_0',PredID1),
   Pred =.. [PredID1,Lex],
   portray_clause((Phon ---> Pred)).

%%%lex_class/2 takes the id for the lexical fsa, obtains its canonical
%%%(term) representation, and calls lex_class/3 to handle either creating
%%%a new interaction predicate if necessary, and returning the
%%%interaction predicate identifier.

lex_class(LexID,PredID) :-
   get_canonrep(LexID,CanonRep),
   lex_class(LexID,CanonRep,PredID).

%%%lex_class/3 checks to see if the lexical fsa's canonical
%%%representation has already been asserted.  If so, we will just use the
%%%interaction predicate corresponding to that fsa, and return its
%%%identifier as the identifier for this lexical entry as well.  If not,
%%%we assert the canonical representation, and call make_new_intpred/2 to
%%%create a new interaction predicate and return the predicate identifier.

lex_class(LexID,CanonRep,PredID) :-
   (   canonrep(ClassID,CanonRep)
   ->  retract(lex_info(Phon,LexID,Lex)),
       assert(lex_info(Phon,ClassID,Lex)),
       concat(q,ClassID,PredID),
       retractall(lex_arc(LexID,_,_,_))
   ;
       assert(canonrep(LexID,CanonRep)),
       make_new_intpred(LexID,PredID)
   ).

%%%make_new_intpred/2 uses concat/2 to obtain the interaction predicate
%%%functor name.  It then asserts this interaction predicate as a
%%%functional relation, and licenses the base lexical entry.  Finally, it
%%%calls do_qrep/2 to define the rest of the predicate, and retracts all
%%%final_state/1 clauses.

make_new_intpred(LexID,PredID0) :-
   concat(q,LexID,PredID0),
   format('~n~n% Lexical class "~p"~n',[PredID0]),
   concat(PredID0,'_0',PredID1),
   FunPred =.. [PredID1,+,-],
   portray_clause((fun FunPred)),
   Pred =.. [PredID1,Var,Var],
   portray_clause((Pred if true)),
   do_qrep(LexID,PredID0),
   retractall(final_state(_State)).

%%%do_qrep/2 is a failure-driven loop.  For every arc in the lexical fsa,
%%%it asserts an 'if' clause licensing further derivations from this node
%%%in the fsa, and calls make_final_state/2 to take care of licensing the
%%%output of this particular lexical rule application.

do_qrep(LexID,PredID0) :-
   lex_arc(LexID,StartState,EndState,RuleID),
   concat(PredID0,'_',PredID1),
   concat(PredID1,StartState,Functor1),
   concat(PredID1,EndState,Functor2),
   Pred1 =.. [Functor1,In,Out],
   Pred2 =.. [Functor2,Aux,Out],
   LexRulePred =.. [RuleID,In,Aux], 
   portray_clause((Pred1 if LexRulePred, Pred2)),
   make_final_state(PredID0,EndState),
   fail
   ;
   nl,true.

%%%make_final_state/2 checks to see if this state is already licensed as
%%%a final state.  If not, it asserts a final_state/1 clause, and writes
%%%the necessary clause of the interaction predicate.

make_final_state(PredID0,State) :-
   final_state(State)
   ->
   true
   ;
   assert(final_state(State)),
   concat(PredID0,'_',PredID1),
   concat(PredID1,State,PredID2),
   Pred =.. [PredID2,Var,Var],
   portray_clause((Pred if true)).

%%%get_canonrep/2 takes a lexical ID and translates the automaton
%%%associated with that lexical entry into a compound term
%%%representation, fsa/2.  fsa/2 includes only state names.

% fsa(start_state,follow_list)
% follow list comprised of cycle/1 and arc/2.
% cycle(end_state)
% arc(end_state,follow_list)

get_canonrep(LexID,fsa(0,Transitions)) :-
   get_transitions(LexID,0,Transitions).

%%%get_transitions/3 takes an ID and a state, and returns the list
%%%of  
get_transitions(LexID,StartState,Transitions) :-
   findall(EndState,lex_arc(LexID,StartState,EndState,_RuleID),EndStates),
   get_fsa(EndStates,LexID,StartState,Transitions).

get_fsa([],_LexID,_StartState,[]).
get_fsa([EndState|Rest],LexID,StartState,[cycle(EndState)|Tail]) :-
   StartState >= EndState,!,
   get_fsa(Rest,LexID,StartState,Tail).
get_fsa([EndState|Rest],LexID,StartState,[arc(EndState,Transitions)|Tail]) :-
   StartState < EndState,!,
   get_transitions(LexID,EndState,Transitions),
   get_fsa(Rest,LexID,StartState,Tail).
