% $Id: step_one.pl,v 1.16 2003/11/15 05:17:58 dm Exp $

:- op(1200,xfx,'lex_rule').
:- op(1100,xfx,'===').

:- dynamic rule_id/1.
:- dynamic lex_rule_desc/3.
:- dynamic frame/3.
:- dynamic lex_rule/3.
:- dynamic they_unify/0.
%-----------------------------------------------------------------------

% step_one/0
% does the necessary retracts and creates a unique ID for each lex_rule
% Then it creates the frame predicates for the lexical rule

step_one :-
   initialize_step_one,
   fail_if_undefined((Name === Input lex_rule Output)),
   do_frames(Input,Output,Frames),
   retract(rule_id(Num)), 
   ID is Num + 1,
   lr_assert(rule_id(ID)),
   my_concat(Name,ID,NewID),
   create_preds(NewID,Input,Output,Frames),
   fail
   ;
   true.

initialize_step_one :-
   retractall(lex_rule_desc(_,_,_)),
   retractall(frame(_,_,_)),
   retractall(lex_rule(_,_,_)),
   retractall(rule_id(_)),
   assert(rule_id(0)).



% create_preds(+ID,+Input,+Output,+Frames)
% creates both description-level and internal-level (feature structure)
% frame and lexical rule predicates

create_preds(ID,Input,Output,Frames) :-
    lr_assert(lex_rule_desc(ID,Input,Output)),
    desc2fs(Input,InFS),
    desc2fs(Output,OutFS),
    concat(frame_,ID,FrameName),
    FramePred =.. [FrameName,InVar,OutVar], 
    LexRulePred =.. [ID,(Input,InVar), (Output,OutVar)], 
    format('~n% Lexical rule "~p"~n',[ID]),
    portray_clause((LexRulePred if FramePred)),
    lr_assert((lex_rule(ID,InFS,OutFS) :- frame(ID,InFS,OutFS))),
    create_list_frames(Frames,ID).

% create_list_frames(FrameList,ID)
% creates as many frames as are in the lists, where the lists run parallel:
% the head of InList is matched with the head of OutList.
% We do _not_  use binrels any more.

create_list_frames([],_).
create_list_frames([In-Out|Rest],ID) :-
   lr_assert(frame(ID,In,Out)), % creates an individual frame.
   make_framedesc(In,Out,ID),
   create_list_frames(Rest,ID).

make_framedesc(InFS,OutFS,ID) :-
   desc2fs([_,_],ListFS),
   feat_fs2featval(hd,ListFS,Hd1),
   ud(InFS,Hd1),
   path_fs2pathval([tl,hd],ListFS,Hd2),
   ud(OutFS,Hd2),
   fs2desc(ListFS,(hd:InDesc,tl:(hd:OutDesc,tl:_))),
   concat(frame_,ID,FrameName),
   FramePred =.. [FrameName,InDesc,OutDesc], 
   portray_clause((FramePred if true)).

%%%make_framedesc(In,Out,ID) :-
%%%   desc2fs(binrel,Binrel),
%%%   feat_fs2featval(in,Binrel,InFeat),
%%%   feat_fs2featval(out,Binrel,OutFeat),
%%%   ud(InDesc,In),
%%%   ud(OutDesc,Out),
%%%   fs2desc(Binrel,FrameArg),
%%%   portray_clause(frame((a_ ID), FrameArg)), nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% A frame predicate will be an object of type binrel, which has two
% features: in and out, whose values are objects of type word.  We
% will call the value of in In, and the value of out Out.  
% Information is passed from the input to a lexical rule to its
% output this way:
% Where we want path equalities for some feature, both In and Out
% are given the same variable as the value of that feature.  
% In is unified with the lr Input, Out is unified with the lr 
% Output, and so the specification of that feature value is passed
% from Input to Output.

% Something else need to be said about the types of In and Out--
% in other words, what input-output type combinations should we
% license.  Normally, the type of the input and output should be
% the same, but under some conditions this is not the case.

% We deal with types by only using species (maximally specific types)
% to calculate frames.  We obtain the list of maximally specific
% subtypes (or species) of the input and output types, and then 
% use an algorithm which depends on the relationship between the 
% input and output types to determine for which subtypes/species we
% want to license the application of the lexical rule.

% That is, we calculate a frame for every
% desired species pair.  The first member of the pair is a species
% of the input type, and the second is a species of the output type.

% General idea: given a lexical rule input and output descriptions, 
% calculates frame per possible input feature structure (given
% disjunctions in the input).  

% Given a particular input FS, we derive a  list of
% disjunction-less output descriptions from given output desc,
% get the corresponding output FS, and then for each output 
% Description-FS pair, calculate the species pairs for which 
% we want frame predicates.  Then, for a particular species pair...

% We create "seed" FSs on the basis of the In and Out types that
% constitute the species pair, and get a list of appropriate features
% for the Out type.  We then walk through the output description,
% removing mentioned features from the approp. feats list.  

% (When we
% find a feature with a non-atomic value, we go through the whole
% process with the description which is the value of that feature in the 
% output, calculating a list of "small" frames, which are incorporated
% into the larger one).

% At the top level--once we have walked through the entire output 
% description, path equalities are added for features that
% remain on the list, and are also appropriate for the input.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%do_frames/3 takes the input and output descriptions of a lexical rule and
% returns the list of frame predicates.  It first collects a list of 
% input FS corresponding to the description (thereby handling disjunctions
% in the input).  It then calls get_outputPairs/2 with the output desc,
% which returns a list of disjunction-less Desc-FS pairs (thereby handling
% disjunctions in the output.  Finally, it passes the list of input FSs,
% and the list of output pairs to frames_byInput/4, which returns a 
% list of frame predicates.

do_frames(InputDesc,OutputDesc,FrameList):-
    findall(InputFS,desc2fs(InputDesc,InputFS),InputList),
    get_outputPairs(OutputDesc,PairList),
    frames_byInput(InputList,PairList,FrameList,[]).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%frames_byInput/4 recursively calculates the frame predicates for each
% input FS on the list, by calling frames_allOutputs/4, which takes one
% input FS and the list of output Desc-FS pairs, and returns a list of
% frame predicates.  

frames_byInput([],_,X,X).
frames_byInput([InputFS|Inputs],PairList,FrameList,Tail) :-
    frames_allOutputs(PairList,InputFS,FrameList,FrameList1),
    frames_byInput(Inputs,PairList,FrameList1,Tail).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% frames_allOutputs/4 recursively works through the list of output 
% Desc-FS pairs, calling get_framelist/5 with the output desc and 
% FS, and input FS, which returns a list of frame predicates.  It
% appends this list to the list returned by the recursive call to 
% frames_allOutputs/4 with the rest of the output pair list.

frames_allOutputs([],_InputFS,X,X).
frames_allOutputs([OutputDesc-OutputFS|Pairs],InputFS,FrameList,Tail) :-
    get_frameList(OutputDesc,OutputFS,InputFS,FrameList,FrameList1),
    frames_allOutputs(Pairs,InputFS,FrameList1,Tail).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%get_frameList/5 takes an output description, FS, and input FS and returns 
% a list of frame predicates.  


% In the first case, the output desc is an atomic type, and it returns 
% [] as the list of frame predicates.

get_frameList(OutputDesc,_OutputFS,_InputFS,[],_Tail) :- 
     type(OutputDesc),
     immed_subtypes(OutputDesc,[]),
     approp_feats(OutputDesc,[]), !.


%%%In the second case, the types of both input and output are gotten
%%%from the corresponding FSs, and a list of species pairs is gotten
%%%via a call to types2pairs/3.  This pair-list is fed to
%%%frames_byPairList/6, and the list of frames is returned.

get_frameList(OutputDesc,OutputFS,InputFS,FrameList,Tail):-
    fs2type_feats(OutputFS,OutputType,_),
    fs2type_feats(InputFS,InputType,_),
    types2pairs(InputType,OutputType,PairList),
    frames_byPairList(PairList,OutputDesc,OutputFS,InputFS,FrameList,Tail).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% frames_byPairList/5 recursively works through the list of species-pairs, 
% calling frames_byPair/7 for each one, and appending the resulting list
% of frame predicates to the rest.


frames_byPairList([],_,_,_,X,X).
frames_byPairList([InType-OutType|Pairs],OutputDesc,OutputFS,InputFS,FrameList,Tail):-
    frames_byPair(InType,OutType,OutputDesc,OutputFS,InputFS,FrameList,FrameList1),
    frames_byPairList(Pairs,OutputDesc,OutputFS,InputFS,FrameList1,Tail).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% frames_byPair/7 Takes the In type and the Out type (that is, what was the 
% species pair) and makes skeleton FSs out of them.  
% It then gets the list of appropriate features for the Out type, and calls 
% get_frames/7, which has the core frame list [In-Out] as the FramesIn 
% argument, and the list of appropriate features as the FeatsIn argument. 
% get_frames/7 will return a list of unmentioned features as the Feats
% argument, and a list of frames to which path equalities should be added
% for those unmentioned features.  This is done via a call to 
% add_patheqsList/2, which takes Feats and FrameList as arguments. 

frames_byPair(InType,OutType,OutputDesc,OutputFS,InputFS,FrameListOut,FrameListIn) :-
    desc2fs(InType,In),
    desc2fs(OutType,Out),
    approp_feats(OutType,AppropFeats),
    get_frames(OutputDesc,OutputFS,InputFS,AppropFeats,Feats,[In-Out],FrameList),
    add_patheqsList(FrameList,Feats),
    append(FrameList,FrameListIn,FrameListOut).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% get_frames/7 walks through the Output description, and removes features
% from the list where appropriate.  It has separtate clauses for 
% conjunction, Feat:Val, and an else clause.  A disjunction clause
% is not needed, because disjunctions are removed earlier. 

%---------------------------------------------------------------------
% In the case of a conjunction, get_frames/8 calls itself for each 
% conjunct.  The frame-list and feat-list outputs for the first 
% conjunct are the inputs for the second, whose outputs are also the 
% overall outputs. 

get_frames((Desc1,Desc2),OutputFS,InputFS,FeatsIn,FeatsOut,FramesIn,FramesOut):- !,
    get_frames(Desc1,OutputFS,InputFS,FeatsIn,Feats,FramesIn,Frames),
    get_frames(Desc2,OutputFS,InputFS,Feats,FeatsOut,Frames,FramesOut).

%---------------------------------------------------------------------
% In the case of Feat:Val, we want to go through the frame-calculation
% process on the basis of the description corresponding to Val.  That is,
% we want a list of "small" frames (calculated on the basis of species
% pairs just as at the top level).  The resulting output list of "big" 
% frames should be the result of getting every possible combination 
% of big frames on the input list with some small frame as its value
% for Feat.  If, on the other hand, Val is atomic, then nothing should
% be done at all to the big frames, since no path equalities are necessary
% below Feat.

% Specifically, we get the small FSs corresponding to the value of Feat
% in the big input and output FSs.  These are our new input and output 
% FS arguments.  We pass these, along with Desc, to get_frameList/5, 
% which returns the small frame list.  We then remove Feat from the list
% of appropriate features, and the resulting list is the FeatsOut 
% argument.  If the small frame list is empty (that is, the value of 
% Feat is atomic) then the frame list input is also the output.  
% Otherwise, we call incorp_smallFSList, which gets all the possible
% big/small frame combinations.
 
get_frames((Feat:Desc),OutputFS,InputFS,FeatsIn,FeatsOut,FramesIn,FramesOut):-
   !,
   (   var(Desc)		% Is this a reentrancy?
   ->  SmallFrameList = []
   ;   feat_fs2featval(Feat,InputFS,SmallInputFS),
       feat_fs2featval(Feat,OutputFS,SmallOutputFS),
       get_frameList(Desc,SmallOutputFS,SmallInputFS,SmallFrameList,[])
   ),
   subtractFeat_list(FeatsIn,Feat,FeatsOut),
   (    SmallFrameList == [] 
   ->   FramesIn = FramesOut
   ;    incorp_smallFSList(FramesIn,Feat,SmallFrameList,FramesOut)
   ).

%---------------------------------------------------------------------
% If the description is anything but a conjunction or a Feat:Val, (a type,
% a variable), we don't take anything off the feats list, and we do 
% nothing to the frame list.

get_frames(_,_,_,Feats,Feats,InOutList,InOutList).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%5


% incorp_smallFSList/4 recursively works through a list of frames, 
% and for each, calculates all possible resulting frames with some
% small frame as the value of Feat.  

% Given a particular In-Out pair, it uses findall to get the list of 
% In-Out pairs which are just like the original, except that they 
% each have a different member of the small frame list (SmallIn-SmallOut) 
% as their values for Feat.  This is accomplished by unifying SmallIn 
% with the value of Feat in In, and unifying Small-Out with the value of 
% Feat in Out.  The resulting list is appended to the framelist output
% for the rest of the framelist input.

% # try to make tail recursive(using difference list arguments)

incorp_smallFSList([],_,_,[]).
incorp_smallFSList([In-Out|Frames],Feat,SmallFrameList,FramesOut) :-
    findall(In-Out,(member(SmallIn-SmallOut,SmallFrameList),
                    feat_fs2featval(Feat,In,InVal),
                    feat_fs2featval(Feat,Out,OutVal),
                    ud(SmallIn,InVal),
                    ud(SmallOut,OutVal)),FrameList1),
    incorp_smallFSList(Frames,Feat,SmallFrameList,FrameList2),
    append(FrameList1,FrameList2,FramesOut).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% All the predicates you need for getting species-pairs lists
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%This is the non-flattening inerpretation from Meurers (2001).

% types2pairs/3 takes the input and output types, obtains their
% species lists from speciesList/2, and then gets the list of species
% pairs from species2pairs/4.  This pair list is its output.

types2pairs(InputType,OutputType,PairList) :-
	speciesList(InputType,InputSpecies),
	speciesList(OutputType,OutputSpecies),
	species2pairs(InputSpecies,OutputSpecies,PairList,[]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% species2pairs/4 takes a list of input species, a list of output
% species, and returns a list of species pairs.

species2pairs([],_OutputSpecies,Tail,Tail).

% In the case that
% an input species X-X is a member of the output species list,
% only X-X is added to the list of pairs.
% Otherwise, all X-Y are added to the species pair list, where
% Y is some output species.

species2pairs([X|InputSpecies],OutputSpecies,PairList,Tail) :-
   (   member(X,OutputSpecies)
   ->  PairList = [X-X|PairList1],
       species2pairs(InputSpecies,OutputSpecies,PairList1,Tail)
   ;   pairs_forOne(OutputSpecies,X,PairList,PairList1),
       species2pairs(InputSpecies,OutputSpecies,PairList1,Tail)
   ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pairs_forOne/4 takes a single input species X and the list of output
% species, creating the list of all X-Y, Y an output species.

pairs_forOne([],_X,Tail,Tail) :- !.
pairs_forOne([Y|OutputSpecies],X,[X-Y|Rest],Tail) :-
    pairs_forOne(OutputSpecies,X,Rest,Tail).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%speciesList/2 takes a type and returns a list of its 
% maximally specific subtypes, or species.  If the type has
% no immediate subtypes, it returns the singleton list with 
% itself as the only member.  Otherwise, it calls handle_subtypes/2
% with the list of immediate subtypes, which returns the species
% list.

speciesList(Type,SortedSpeciesList) :-
    speciesList(Type,SpeciesList,[]),
    sort(SpeciesList,SortedSpeciesList).

speciesList(Type,SpeciesList,SpeciesTail) :-
   immed_subtypes(Type,Subtypes),
   (  Subtypes == []
   -> SpeciesList = [Type|SpeciesTail]
   ;  handleTypes(Subtypes,SpeciesList,SpeciesTail)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% handle_subtypes/2 takes a list of subtypes and recursively works
% through it, ultimately returning a list of their maximally
% specific subtypes, via depth-first search.  If the head of the
% list is itself a species (that is get_speciesList/2 returns
% [Subtype], then Subtype is the head of the output list, where
% the rest of the list is gotten by a recursive call to 
% handle_subtypes/2 with the rest of the input list as its input.

% Otherwise, a list of immediate subtypes is obtained for the head
% of the list, appended to the rest of the input list, and the 
% resulting list is input to a recursive call to handle_subtypes/2.

handleTypes([],X,X).
handleTypes([H|T],SpeciesList,SpeciesTail):-
   speciesList(H,SpeciesList,NewTail),
   handleTypes(T,NewTail,SpeciesTail).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% end of species-pairs predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% get_outputPairs/2 takes an output description which may contain
%%%% disjunctions and returns the list of disjunction-free description-
%%%% FS pairs.  It first gets a list of disjunction-free descriptions
%%%% with a call to disjuncts_out/2, then gets the list of corresponding
%%%% description-FS pairs with a call to desc2_DescFS/2.

get_outputPairs(OutputDesc,PairList) :-
    disjuncts_out(OutputDesc,DescList),
    desc2_DescFS(DescList,PairList).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% desc2_DescFS/2 takes a list of descriptions, translates each in 
% turn to a FS, and returns the list of Desc-FS pairs.
% version 2 -- to deal with desc2fs/2 altering Prolog variables

desc2_DescFS([],[]).
desc2_DescFS([Desc|Descs],[Desc-FS|Rest]) :-
    copy_term(Desc,Desc2),
    desc2fs(Desc2,FS),
    desc2_DescFS(Descs,Rest).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% add_patheqsList/2 takes a list of frames and a list of features
% and works through the list of frames, calling add_patheqs/2 for 
% each, which adds path equalities for every feature on the feature
% list.

add_patheqsList([],_).
add_patheqsList([Frame|Frames],FeatList) :-
    add_patheqs(FeatList,Frame),
    add_patheqsList(Frames,FeatList).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% add_patheqs/2 takes a frame In-Out and a list of features 
% that are appropriate to the Out type of the In-Out pair and 
% works through the list of features, adding a path equality for
% every one that is also appropriate for the In type. 
 
% Specifically, add_patheqs/2 takes a feature Feat, gets the type
% of In, and if Feat is appropriate to In, it gets the value of 
% Feat for both In and Out, and unifies a variable with both, thereby
% enforcing a path equality.

add_patheqs([],_).

add_patheqs([Feat|Rest],In-Out) :- 
     fs2type_feats(In,InType,_),
     (    approp(Feat,InType,_) ->
          feat_fs2featval(Feat,In,SmallIn),
          feat_fs2featval(Feat,Out,SmallOut),
          ud(Val-bot,SmallIn),
          ud(Val-bot,SmallOut)
     ;    true
     ),
     add_patheqs(Rest,In-Out).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% subtractFeat_list/3 just takes a feature Feat and a list of features
% (of which Feat must be a member or the predicate fails -- but this 
% should not be a problem) and succeeds when it find Feat on the list, 
% returning the list less Feat. 

subtractFeat_list([],_,[]).
subtractFeat_list([Feat|Rest],Feat,Rest) :- !.
subtractFeat_list([X|Rest1],Feat,[X|Rest2]) :-
    subtractFeat_list(Rest1,Feat,Rest2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disjuncts_out(Desc,DescList) :-
    rm_disjuncts(Desc,[],DescList).

% rm_disjuncts/3 walks through a description DESC which may contain
% disjunction, and returns a list of disjunctionless descriptions.
% The first argument is the description input.  The second argument
% is the list of disjunctionless descriptions that have been created
% so far.  The third argument is the output, the new list of
% disjunctionless descriptions that results from whatever will be
% done with the input description.

% In the first case, DESC is a variable or a type.  If there are no
% descriptions to add this to, it returns the singleton list with
% DESC as a member.  Otherwise, it calls add_desc2list to add DESC
% to all of the input descriptions.

rm_disjuncts(Desc,DescListIn,DescListOut) :-
   (    var(Desc)
   ;    type(Desc)
   ), 
   !,
   ( DescListIn == []
   -> DescListOut = [Desc]
   ;  add_desc2list(DescListIn,Desc,DescListOut,[])
   ).
    

% In the second case, the input description is a disjunction of
% Desc1 and Desc2.  In this case, rm_disjuncts3 is called twice,
% with each of the disjuncts as an input, where both calls have the
% same input list as their second argument.  This effectively doubles
% the number of descriptions.  The resulting output lists are
% appended.

rm_disjuncts((Desc1;Desc2),DescListIn,DescListOut) :- 
   !,
   rm_disjuncts(Desc1,DescListIn,DescListOut1),
   rm_disjuncts(Desc2,DescListIn,DescListOut2),
   append(DescListOut1,DescListOut2,DescListOut).

% In the third case, the input description is a conjunction of Desc1
% and Desc2.  Here also we have two calls to rm_disjuncts/3, the first
% taking Desc1 and the input list as inputs.  The output of this call
% will be the input list of the second call, taking Desc2 as its first
% argument.  Its output is the overall output.

rm_disjuncts((Desc1,Desc2),DescListIn,DescListOut) :- 
   !,
   rm_disjuncts(Desc1,DescListIn,DescList),
   rm_disjuncts(Desc2,DescList,DescListOut).

% In the fourth case, the input description is of the form Feat:Val.
% We call rm_disjuncts for with Val as input, in order to obtain a
% list of "small" disjunctionless descriptions.  If the input list is
% empty, we just want to return a list of Feat:Vals.  The call to
% feat_vals/3 obtains this list.

% Otherwise, we want to add Feat:Val (for all the possible Vals) to
% the input list. This is done via a call to add_feat2list/5, which
% takes Feat and the SmallDescs list as input, and returns the overall
% output list.

rm_disjuncts((Feat:Desc),DescListIn,DescListOut) :-
   !,
    rm_disjuncts(Desc,[],SmallDescs),
    ( DescListIn == []
    -> featvals(SmallDescs,Feat,DescListOut)
    ;  add_feat2list(SmallDescs,Feat,DescListIn,DescListOut,[])
    ).


% Two special cases for removing special list notation:
rm_disjuncts(X,DescListIn,DescListOut) :-
   X==[],
   !,
   rm_disjuncts(e_list,DescListIn,DescListOut).
rm_disjuncts(X,DescListIn,DescListOut) :-
   functor(X,'.',_),
   X=[H|T],
   !,
   rm_disjuncts((hd:H,tl:T),DescListIn,DescListOut).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% add_desc2list/4 takes a list of descriptions and a single
% descriptions, and adds the single description as a conjunct
% to each of the descriptions on the list.  

add_desc2list([],_,X,X).
add_desc2list([Desc1|Descs],Desc2,[(Desc1,Desc2)|Rest],Tail) :-
    add_desc2list(Descs,Desc2,Rest,Tail).

%--------------------
%add_feat2list(SmallDescs,Feat,DescListIn,DescListOut,Rest) takes a list of
% descriptions as its first argument, and a feature as its second.
% The first list are all possible values of Feat.  add_feat2list/5
% Take a feature as its second argument, a list of possible values
% for that feature as its first argument, and a list of descriptions
% as its third argument.  It peels of the first Desc (=Val), and
% calls add_desc2list/4 with Feat:Desc and the list of descriptions as
% inputs.  add_desc2list/4 will return the list of descriptions with
% Feat:Desc conjoined.  add_feat2list/5 recursively works through the
% list of featvals, ultimately returning a list on which every
% (Desc,Feat:Desc) combination is a member.

add_feat2list([],_,_,X,X).
add_feat2list([Desc|Descs],Feat,DescListIn,DescListOut,Rest) :-
    add_desc2list(DescListIn,(Feat:Desc),DescListOut,DescListTail),
    add_feat2list(Descs,Feat,DescListIn,DescListTail,Rest).
%-----------------------------------------------------------------------
%featvals(SmallDescs,Feat,DescListOut) takes a list of Descs and a
% Feat, and returns a list whose members are of the form Feat:Desc.
featvals([],_Feat,[]).
featvals([Desc|Descs],Feat,[Feat:Desc|Rest]) :-
    featvals(Descs,Feat,Rest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

test_unify(X,Y) :-
	retractall(they_unify),
        unify_aux(X,Y),
	they_unify.

unify_aux(X,Y) :-
	ud(X,Y),
	assert(they_unify),
	fail.
unify_aux(_,_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% my_concat( +Atom1, +Atom2, -Concatenation)

my_concat(A,B,AB):-
     name(A,NA),
     name(B,NB),
     append(NA,NB,NAB),
     name(AB,NAB).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Rewrites Prolog variable as alphabetic variants.

writealph(Term) :-
   (   numbervars(Term,0,_),
       portray_clause(Term),
       fail
    ;  true
    ).
