:- dynamic lex_arc/4.
:- dynamic lexID/1.
:- dynamic lex_info/3.

%%%Step three takes each lexical entry, and builds a fsa for that
%%%lexical entry, based on the (pruned) global fsa which is the
%%%output of step two.

%%%Arc representation:
%%%lex_arc(LexID,StartState,EndState,RuleID)

%%%Unlike the last step, where we are pruning an fsa we have already
%%%built, here we are deciding which arcs of the global fsa we would
%%%like to have in our lexical version.  We save the feature structures
%%%associated with a state in a table, if there are cycles terminating
%%%at that state.

%%%So, in terms of cyclicity, when we arrive at a state we have already
%%%been to, we need to know whether the FS we got there with necessitates
%%%the inclusion of any more arcs along any path leaving the present
%%%state.  We never want to throw away an arc once we already have it,
%%%nor do we want to throw away any FS associated with a particular arc,
%%%for subsumption-checking purposes.  Therefore, table(state,_) may
%%%be multiply defined.

%%%Therefore, if the FS is _not_ subsumed by any of the FSs associated
%%%with that state already, we start the whole process over, asserting
%%%arcs only if they have not already been asserted.


step_three :-
   init_step_three,
   step_three_aux.

init_step_three :-
   retractall(lex_arc(_,_,_,_)),
   retractall(table(_,_)),
   retractall(lexID(_)),
   retractall(lex_info(_,_,_)),
   assert(lexID(0)).

%%%For each lexical entry, step_three_aux/0 assigns a unique
%%%lexical identifier, obtains the internal representation of
%%%the lexical entry description, and calls make_lexfsa/2
%%%with the ID and that FS as arguments.  lex_info/3 is a way
%%%to look up the lexical entry FS and its unique identifier.
%%%We use this in step_four.pl.  Also needed for viewer to
%%%connect Phon value to automaton.
step_three_aux :-
   fail_if_undefined((Phon ---> Lex)),
   (lr_debug -> debug_file(S),portray_clause(S,Phon);true),
   get_lex_id(LexID),
   desc2fs(Lex,LexFS),
   assert(lex_info(Phon,LexID,Lex)),
   make_lexfsa(LexID,LexFS),
   fail
   ;
   true.
	      
%%%For each initial arc and its associated lexical rule,
%%%make_lexfsa/2 makes a copy of the lexical FS (for
%%%purposes of propogation down a particular path),
%%%and checks to see if it unifies with the input
%%%specifications of the lexical rule.  If so, a table
%%%entry is asserted if necessary (as in step_two), and a
%%%lexical fsa arc is asserted with all the same state and
%%%rule information as the corresponding arc in the global
%%%fsa, but with a lexical ID argument added.  Finally,
%%%make_lexfsa/3 is called with the end state, output of
%%%the lexical rule application and the lexical ID as
%%%arguments.  
make_lexfsa(LexID,LexFS) :-
   arc(0,EndState,RuleID),
   copy_term(LexFS,InputFS),
   lex_rule(RuleID,InputFS,OutputFS),
   make_table(EndState,OutputFS),
   lr_assert(lex_arc(LexID,0,EndState,RuleID)),
   make_lexfsa(EndState,OutputFS,LexID),
   fail
   ;
   true.


%%%For a particular state, and an associated FS,
%%%make_lexfsa/3 makes a copy of the input FS and
%%%propagates specifications down all paths beginning
%%%at that state by calling make_lexfsa/5 for every
%%%arc with the input state as its start state.
make_lexfsa(StartState,OrigFS,LexID) :-
   arc(StartState,EndState,RuleID),
   copy_term(OrigFS,FS),
   make_lexfsa(StartState,EndState,RuleID,FS,LexID),
   fail
   ;
   true.

%%%make_lexfsa/5 takes as arguments the start and end
%%%states of a particular arc, that arc's associated
%%%lexical rule ID, a FS associated with the start state,
%%%and the lexical ID.  There are two subcases: the arc is
%%%a cycle, or not.  As noted in step_two.pl, cyclicity is
%%%easy to check for because of the way state IDs are
%%%assigned.

%%%In both cases, the first step is to check whether the
%%%input FS unifies with the input specifications of the
%%%lexical rule.  If not, nothing is done.  But if so:

%%%In the first case, the arc is not a cycle.  If
%%%propogation succeeds, a table entry is asserted if
%%%necessary, the arc is asserted (if not already
%%%asserted), and the process is continued for all
%%%arcs leaving this state using the output of lexical
%%%rule application as input.
make_lexfsa(StartState,EndState,RuleID,InputFS,LexID) :-
   EndState > StartState,
   (   lex_rule(RuleID,InputFS,OutputFS) ->
       make_table(EndState,OutputFS),
       make_arc(LexID,StartState,EndState,RuleID),
       make_lexfsa(EndState,OutputFS,LexID)
   ;
       true
   ).

%%%In the second case, the arc is a cycle.  If propogation
%%%succeeds, an arc is asserted (if not already asserted),
%%%and a call to cycle_check with the end state, the output
%%%of lexical rule application and the lexical ID as
%%%arguments takes care of table maintenance and any further
%%%testing, if necessary.
make_lexfsa(StartState,EndState,RuleID,InputFS,LexID) :-
   StartState >= EndState,
   (   lex_rule(RuleID,InputFS,OutputFS) ->
       make_arc(LexID,StartState,EndState,RuleID),
       cycle_check(EndState,OutputFS,LexID)
   ;
       true
   ).

%%%Given a (new) FS associated with a particular state,
%%%cycle_check/3 first checks to see if there is a
%%%preexisting FS associated with that state which is
%%%more general than this new FS.  If so, nothing more
%%%is done.  If not, then all preexisting table entries
%%%whose FSs are subsumed by this new one are retracted,
%%%a new table entry is asserted, and the FS is
%%%propogated down all paths leaving this state,
%%%by means of a call to make_lexfsa/3 with the state,
%%%the FS, and the lexical ID as arguments.
cycle_check(EndState,OutputFS,LexID) :-
   (   more_specific(EndState,OutputFS) ->
       true
   ;
       more_general(EndState,OutputFS),
       assert(table(EndState,OutputFS)),
       make_lexfsa(EndState,OutputFS,LexID)
   ).

%%%make_arc/4 asserts a lexical fsa arc if an identical
%%%arc has not already been asserted.
make_arc(LexID,StartState,EndState,RuleID) :-
   (   lex_arc(LexID,StartState,EndState,RuleID) ->
       true
   ;
       lr_assert(lex_arc(LexID,StartState,EndState,RuleID))
   ).

%%%get_lex_id/1 retracts the old lexical ID, asserting and
%%%returning the next.
get_lex_id(LexID) :-
   retract(lexID(Num)),
   LexID is Num + 1,
   assert(lexID(LexID)).
