:- ensure_loaded(lr_compiler(follow)).

:- dynamic follow/2.
:- dynamic state_id/1.
:- dynamic arc/3.
:- dynamic retract_arc/2.
:- dynamic table/2.
:- dynamic temp/2. %to be dumped when we have a better version of follows

:- use_module(library(lists),[member/2]).

% step_two/0.  This step has three sub-parts:
%%%1.  Obtain the follows relation.  The follows relations
%%%associates a lexical rule ID with thelist of lexical rule
%%%IDs whose input specifications will unify with the output
%%%specifications of the first.

%%%2.  Based on the follows relation, build a global FSA
%%%encoding all possible combinations of lexical rule application.

%%%3.  Propogation of specifications.  Unify outputs and inputs
%%%along all paths in the global fsa, pruning those arcs that
%%%turn out to be unusable.

%%%State-based arc representation:
%%%arc(StartState,EndState,LexRuleID)

%%%From the start state, every lexical rule has an initial arc.
%%%From that arc's end state, every lexical rule on the follow
%%%list has an arc with that end state as its start state.
%%%The end state will either be the next state, or a state
%%%further back on the path. That is:

%%%While building the fsa, we keep track of the prefix
%%%(a list of endstate-rule pairs that constitutes the path
%%%from the start state to the current node); if a lexical rule
%%%on the follow list has already been seen, we create a cycle
%%%back to the endstate of that prior rule application.

%%%step_two_gen :-
%%%   init_step_two,
%%%   step_two_aux_gen
%%%   fail
%%%   ;
%%%   true.

step_two :-
   init_step_two,
   step_two_aux.

init_step_two :-
   retractall(temp(_,_)),
   retractall(table(_,_)),
   retractall(state_id(_)),
   retractall(arc(_,_,_)),
   assert(state_id(0)).

%%%We currently have two versions of the code for
%%%propogation of specifications, depending on whether
%%%we are using generalized frame predicates, or multiple
%%%definitions of the frame predicate, to test the
%%%unification of output and input specifications.
%%%In the first case, there will be only one (generalized)
%%%FS associated with a node; in the second, there may be
%%%multiple FSs.  Currently, we call the code using
%%%multiple definitions of the frames.

step_two_aux_gen :-
   get_follows,
   construct_automaton,
   prop_specs_gen.
step_two_aux :-
   get_follows,
   construct_automaton,
   prop_specs_multi.


%Asserts the start-state, creates an initial arc for each
%lexical rule, and then the non-initial arcs based on the
%follows relation.
construct_automaton :-
   initial_arcs,
   non_initial_arcs.

%Creates an initial arc for each lexical rule.
initial_arcs :-
   follow(RuleID,_),
   get_arcid(EndState),
   lr_assert(arc(0,EndState,RuleID)),
   fail
   ;
   true.

%%%Failure-driven loop.  For every initial arc, gets the follow
%%%list for the relevant lexical rule, then calls a
%%%predicate to make arcs for every rule on that list.
non_initial_arcs :-
   arc(0,EndState,RuleID),
   follow(RuleID,List),
   make_arcs(List,EndState,[RuleID-EndState]),
   fail
   ;
   true.

%%%make_arcs/3 works through the follow list.  Every arc will
%%%have the same start state.  For a particular rule, checks
%%%to see if it occurs in the prefix.  If so, the prefix list
%%%endstate will also be the endstate for this arc.  If not,
%%%the endstate will be the current id number.  Further, since
%%%this is not a cycle, the follow list for the rule is
%%%obtained and make_arcs/3 is called to handle creating all
%%%the arcs leaving this new state.

make_arcs([],_StartState,_Prefix).
make_arcs([RuleID|Rest],StartState,Prefix) :-
   (   member(RuleID-EndState,Prefix) ->
       lr_assert(arc(StartState,EndState,RuleID))
   ;   get_arcid(EndState),
       lr_assert(arc(StartState,EndState,RuleID)),
       follow(RuleID,List),
       make_arcs(List,EndState,[RuleID-EndState|Prefix])
   ),
   make_arcs(Rest,StartState,Prefix).

%%%get_arcid/1 retracts the old state id (which has already
%%%been used) adds 1 to it, and returns this new id.
get_arcid(ID) :-
   retract(state_id(Num)),
   ID is Num + 1,
   assert(state_id(ID)).

%%%%%%%%%%%Propogation of specifications

%%%Cyclic application of lexical rules may produce a more
%%%general FS (or incompatible FS?), which could propagate
%%%down paths that would have been pruned in an earlier round
%%%of specification propogation. So, we use a data structure
%%%retract_arc/3 to keep track of those arcs which will be
%%%retracted at the end.  If a more general FS is found and
%%%we want to "save" an arc that would otherwise have been
%%%retracted, retract the retract_arc.  We use table(State,FS)
%%%to keep track of FSs for subsumption checks.  However,
%%%since such checking is only done for states which are the
%%%endpoint of a cycle, we assert table entries only for those
%%%nodes.
%-----------------------------------------------
%%%Version using generalized frames:

%%%prop_spec_gen/0 is a failure-driven loop, taking each
%%%initial arc in turn.  It obtains the output FS for the
%%%application of that lexical rule, calls the predicate to
%%%assert a table entry if necessary, then calls
%%%prop_specs_gen/2 to handle all subsequent arcs. All table
%%%entries are retracted when this initial arc is done.

%%%When all initial arcs have been dealt with, arcs
%%%corresponding to all (surviving) retract_arc/2 clauses
%%%are actually retracted, thereby pruning the fsa of unusable
%%%arcs.

prop_specs_gen :-
   arc(0,EndState,RuleID),
   gen_lex_rule(RuleID,_InputFS,OutputFS),
   make_table(EndState,OutputFS),
   prop_specs_gen(EndState,OutputFS),
   retractall(table(_,_)),
   fail
   ;
   retract_arcs,
   (lr_debug -> debug_file(S),write_arcs(S) ; true).


%%%For a particular state, and an associated FS,
%%%prop_specs_gen/2 propagates specifications down all
%%%paths beginning at that state, calling prop_specs_gen/4
%%%for each arc originating at that state.  Failure-driven
%%%loop.

prop_specs_gen(StartState,FS) :-
   arc(StartState,EndState,RuleID),
   prop_specs_gen(StartState,EndState,RuleID,FS),
   fail
   ;
   true.

%%%prop_specs_gen/4 tests to see if the input FS it is given
%%%will unify with the input specifications of the lexical
%%%rule associated with the start and end states it is given.
%%%There are two cases, depending on whether this arc is a
%%%cycle or not.  Cyclicity is easy to detect, since state
%%%names are integers.  If an arc is not the endpoint of a
%%%cycle, the endstate id will be greater than the startstate
%%%id.  If it is the endpoint of a cycle, the startstate id
%%%will be greater than or equal to the endstate id.

%%%In the first case, not a cycle.  If unification succeeds,
%%%assert a table entry if necessary, take back the death
%%%warrant for this arc if there is one, and propogate
%%%specifications further down all paths leaving this end state.
%%%If not, assert a retract_arc for this arc.
prop_specs_gen(StartState,EndState,RuleID,InputFS) :-
   EndState > StartState,
   (   gen_lex_rule(RuleID,InputFS,OutputFS) ->
       make_table(EndState,OutputFS),
       fix_retracts(StartState,EndState),
       prop_specs_gen(EndState,OutputFS)
   ;
       assert(retract_arc(StartState,EndState))
   ).

%%%In the second case, a cycle.  If propogation succeeds, then
%%%take back death warrant if there is one, and see if the
%%%resulting output FS is more general than (or possibly
%%%incompatible with) any FS already associated with this node.
%%%If so, we want to propogate this FS down all paths leaving
%%%the current state, to see if unification will succeed
%%%anywhere it did not previously. If unification of the
%%%current input with the lexical rule specifications did not
%%%succeed to begin with, we assert a retract_arc.
prop_specs_gen(StartState,EndState,RuleID,InputFS) :-
   StartState >= EndState,
   (   gen_lex_rule(RuleID,InputFS,OutputFS) ->
       fix_retracts(StartState,EndState),
       (   check_table(EndState,OutputFS) ->
	   prop_specs_gen(EndState,OutputFS)
       ;
	   true
       )
   ;
       assert(retract_arc(StartState,EndState))
   ).


%------------------------------------------------------------
%%%Propogation of specifications with multiple frames

%%%prop_spec_multi/0 is a failure-driven loop, taking each
%%%initial arc in turn.  It obtains a list of output FSs for
%%%the application of that lexical rule, calls the predicate
%%%to assert table entries if necessary, then calls
%%%prop_specs_mult/2 to handle all subsequent arcs. All table
%%%entries are retracted when this initial arc is done.

%%%When all initial arcs have been dealt with, arcs
%%%corresponding to all (surviving) retract_arc/2 clauses
%%%are actually retracted, thereby pruning the fsa of unusable
%%%arcs.
prop_specs_multi :-
   arc(0,EndState,RuleID),
   findall(OutputFS,lex_rule(RuleID,_InputFS,OutputFS),Outputs),
   make_table_multi(EndState,Outputs),
   prop_specs_multi(EndState,Outputs),
   retractall(table(_,_)),
   fail
   ;
   retract_arcs,
   (lr_debug -> debug_file(S),write_arcs(S) ; true).

%%%For a particular state, and an associated list of FSs,
%%%prop_specs_multi/2 propagates specifications down all
%%%paths beginning at that state, calling prop_specs_gen/4
%%%for each arc originating at that state.  Failure-driven
%%%loop.
prop_specs_multi(StartState,Outputs) :-
   arc(StartState,EndState,RuleID),
   prop_specs_multi(StartState,EndState,RuleID,Outputs),
   fail
   ;
   true.

%%%Given a list of input FSs, prop_specs_multi/4 tests to
%%%see which, if any, will unify with the input
%%%specifications of the lexical rule associated with
%%%If the resulting list of outputs is empty, then we assert
%%%a retract_arc corresponding to this arc.  If the list
%%%is not empty, we take back the death warrant for this arc
%%%if there is one, and call prop_specs_multi/3 to take care
%%%of table maintenance and further propogation (if necessary).

prop_specs_multi(StartState,EndState,RuleID,OrigInputs) :-
   copy_term(OrigInputs,Inputs),
   findall(Output,(member(Input,Inputs),
		   lex_rule(RuleID,Input,Output)),Outputs),
   (   Outputs == [] ->
       assert(retract_arc(StartState,EndState))
   ;
       fix_retracts(StartState,EndState),
       prop_specs_multi(StartState,EndState,Outputs)
   ).


%%%There are two cases, depending on whether this arc is a
%%%cycle or not.  Cyclicity is easy to detect, since state
%%%names are integers.  If an arc is not the endpoint of a
%%%cycle, the endstate id will be greater than the startstate
%%%id.  If it is the endpoint of a cycle, the startstate id
%%%will be greater than or equal to the endstate id.

%%%In the first case, not a cycle.  Assert a table entry
%%%if necessary, and propogate specifications further
%%%down all paths leaving this end state.
prop_specs_multi(StartState,EndState,Outputs) :-
   EndState > StartState,
   make_table_multi(EndState,Outputs),
   prop_specs_multi(EndState,Outputs).

%%%In the second case, a cycle.  Given the list of outputs, we
%%%remove all those that are more specific than some FS
%%%already associated with the node.  If none are left, we
%%%do no further propogation.  If the list is non-empty,
%%%however, we want to propogate these FSs down all paths leaving
%%%the current state, to see if unification will succeed
%%%anywhere it did not previously.
prop_specs_multi(StartState,EndState,OutputsIn) :-
   StartState >= EndState,
   check_table_multi(OutputsIn,EndState,OutputsOut),
   (   OutputsOut = [] ->
       true
   ;
       prop_specs_multi(EndState,OutputsOut)
   ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxilliary predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%Instead of writing arcs as they are asserted, writes
%%%all that have been asserted when called.
write_arcs(S) :-
   arc(StartState,EndState,RuleID),
   portray_clause(S,arc((a_ StartState),(a_ EndState),(a_ RuleID))),
   fail
   ;
   true.

%------------------------------------------------------------
%%%For every retract_arc/2 clause, retracts the corresponding
%%%arc, and if the arc is not the final arc in a cycle,
%%%retracts all subsequent arcs.
retract_arcs :-
   retract_arc(StartState,EndState),
   retract(arc(StartState,EndState,_RuleID)),
   (   EndState > StartState ->
       prune_unreachables(EndState)
   ;
       true
   ),
   fail
   ;
   true.

%%%For a given state, prunes every arc leaving that state, and
%%%all arcs subsequent to those arcs.
prune_unreachables(State) :-
   retract(arc(State,EndState,_RuleID)),
   prune_unreachables(EndState),
   fail
   ;
   true.

%--------------------------------------------------------------
%%%If there has been a retract_arc/2 clause asserted for a
%%%given start and end state, retracts it.
fix_retracts(StartState,EndState) :-
   retract_arc(StartState,EndState) ->
   retract(retract_arc(StartState,EndState))
   ;
   true.

%--------------------------------------------------------------
%%%For a particular state, if any arc with it as its end state
%%%is the final arc in a cycle, then asserts table entries for
%%%all the FSs on the given list.
make_table_multi(EndState,FSlist) :-
   (   (   arc(StartState,EndState,_RuleID),
	   StartState >= EndState
       )
   ->
       make_table_multi_aux(FSlist,EndState)
   ;
       true
   ).
%%%Runs through the list of FSs on the input list, asserting
%%%each as a table entry associated with the given state.
make_table_multi_aux([],_State).
make_table_multi_aux([FS|Rest],State) :-
   assert(table(State,FS)),
   make_table_multi_aux(Rest,State).

%-------------------------------------------------------------
%%%If the given state is the endstate of the final arc in a
%%%cycle, then assert a table entry associating that state
%%%with the given FS.
make_table(EndState,FS) :-
   (   arc(StartState,EndState,_RuleID),
       StartState >= EndState ->
       assert(table(EndState,FS))
   ;
       true
   ).

%-------------------------------------------------------------
%%%Given a list of FS outputs arriving back at a node
%%%from a cycle, want to see if there are any which
%%%should be propogated down all paths leaving that node.
%%%Two cases: 1) more general than, or 2) incompatible
%%%with (is this a viable case?) other FSs stored
%%%for that node.  If all new outputs are subsumed by
%%%old ones, then we don't want to propogate specs again,
%%%and we don't want to store anything new.  Otherwise,
%%%we want to propogate unsubsumed new outputs, and store
%%%them, throwing away any old FSs subsumed by them.

%%%check_table_multi/3 works through a list of FSs,
%%%discarding those that are more specific than a FS
%%%already associated with that state.  Any old table
%%%entry whose associated FS is subsumed by some new FS
%%%is retracted.  All new FSs which are not discarded are
%%%asserted as table entries.

check_table_multi([],_State,[]).
check_table_multi([FS|Rest1],State,Rest2) :-
   more_specific(State,FS),!,
   check_table_multi(Rest1,State,Rest2).
check_table_multi([FS|Rest1],State,[FS|Rest2]) :-
   more_general(State,FS),
   assert(table(State,FS)),
   check_table_multi(Rest1,State,Rest2).

%%%more_specific/2 takes a state and its associated (new) FS,
%%%and succeeds if it can find an old FS associated with that
%%%state which is more general than the new one.
more_specific(State,NewFS) :-
   table(State,OldFS),
   subsumeFS(OldFS,NewFS).

%%%more_general/2 takes a state and its associated (new) FS, and
%%%retracts all old FSs associated with that state that are more
%%%specific than it.
more_general(State,NewFS) :-
   table(State,OldFS),
   subsumeFS(NewFS,OldFS),
   retract(table(State,OldFS)),
   fail
   ;
   true.
   
   
%--------------------------------------------------------------
%%%some version of this is needed for prop_specs_gen, but this one
%%%isn't right.
%%%check_table(State,NewFS) :-
%%%   more_specific(State,NewFS), !.
%%%check_table(State,NewFS) :-
%%%   more_general(State,FS),
%%%   assert(table(State,FS)),
%%%   fail.

%-------------------------------------------------------------
% TO BE REPLACED BY BETTER CODE
% (WHICH TAKES GENERALIZATION OF FRAMES INTO ACCOUNT)
% get_follows

get_follows :-
    (   lex_rule(ID1,_,_),
	(   lex_rule(ID1,_,Output1),
	    lex_rule(ID2,Input1,_),
	    (   test_unify(Output1,Input1) ->
		(   temp(ID1,ID2) ->
		    true
		;   lr_assert(temp(ID1,ID2))
		)
	    ;   true
	    ),
	    fail
	;   true
	),
        % convert an enumerator into a list here
	get_list(List),
	(   follow(ID1,List) ->
	    true
	;   lr_assert(follow(ID1,List))
	),
        fail
    ;   true
    ).

unify_fs(X,X,X).

get_list([ID2|Rest]) :-
    retract(temp(_,ID2)),
    !,
    get_list(Rest).
get_list([]).

% subsumeFS(+FS1f,FS2)
% gets the types of FS1 and FS2 and sees if FS1's type is more general,
% then checks if the values for the appropriate features of FS1 subsume
% those for FS2

subsumeFS(FS1,FS2) :-
    fs2type_feats(FS1,Type1,_),
    fs2type_feats(FS2,Type2,_),
    unify_type(Type1,Type2,Type2),
    approp_feats(Type1,List),
    assign_somevals(List,FS1,FS2).

% assign_somevals(+FeatList,+FS1,+FS2)
% walks through the FeatList, seeing if the smaller feature structure for
% FS1 subsumes the smaller one for FS2

assign_somevals([],_,_).
assign_somevals([Feat|Rest],FS1,FS2) :-
    feat_fs2featval(Feat,FS1,Val1),
    feat_fs2featval(Feat,FS2,Val2),
    subsumeFS(Val1,Val2),
    assign_somevals(Rest,FS1,FS2).