%%% ==================================================================
%%%
%%% Some extensions of the core TRALE system
%%%
%%% $Id: trale_extras.pl,v 1.79 2003/11/15 00:22:45 dm Exp $
%%%
%%% Time-stamp: <dm on 14.Nov.2003 19:22:11 EST>
%%%
%%% ==================================================================

%%% shortcuts

c :- compile_gram(theory).
dc :- dcompile_gram(theory).    % for use with source-level debugger

% Allow the user to write :- tree_extensions. to get the additional
% syntax for integrating the extralogical parts (phonology/dtrs)
% into the feature structure. Note that this has to be part of
% the grammar file of the user, since it contains parts that need
% to be compiled by the trale grammar compiler.

tree_extensions :- [trale_home(tree_extensions)].

%%% --------------------------------------------------------------- %%
%%% Starting up code----------------------------------------------- %%
%%% --------------------------------------------------------------- %%

:- use_module(library(system),[system/1,delete_file/2]).
:- use_module(library(charsio),[format_to_chars/3]).

%% command which is executed at every startup 
%% (be it from a state or from the prolog source code)

trale_startup :-
   % load user startup file unless it has already been done by loading
   % the prolog code of startup.pl (i.e. when LOAD_SOURCES is false)
   (  (environ('LOAD_SOURCES',false), environ('READ_USER_STARTUP',true))
   -> load_user_startup
   ;  true
   ),
   %
   % connect to graphical interface if needed
   (  (environ('INTERFACE',Interface), 
       (Interface=='graphics(remote)';Interface=='graphics(true)'))
   -> (  fail_if_undefined(start_client)
      -> true
      ;  format(user_error,'~N**Error: startup of interface failed~n',[])
      )
   ;  nl
   ),
   %
   % output program and version header
   trale_milca_version(Version),
   format('~n~nTRALE Milca environment (version ~p)~n',[Version]),
   format('Copyright (C) 2002/3 Project MILCA A4~n',[]),
   format('PIs: Detmar Meurers (OSU), ',[]),
   format('Gerald Penn (Univ. Toronto), ',[]),
   format('Frank Richter (Univ. Tbingen)~n',[]),
   format('All rights reserved~n~n~n',[]),
   %
   % if a $USER_CMD_FILE exists, load it
   (  (environ('USER_CMD_FILE',Name),file_exists(Name))
   -> format('~NCompiling grammar/executing user commands (options -c and -e):~n~n',[]),
      consult(Name)
   ;  true
   ).

%%% --------------------------------------------------------------- %%
%%% Saving of system states --------------------------------------- %%
%%% --------------------------------------------------------------- %%

% top level predicate to save state of currently loaded trale version:
save_trale_state :-
   environ('CORE_NAME',CoreName),
   save_state(trale_home(CoreName)).

% same, but for saving a version of trale, with compiled grammar, for tsdb 
save_tsdb_state :-
   load_tsdb,
   save_state(tsdb,(trale_startup,itsdb:connect_tsdb)).

%% save_state/1: saving user states under name given as argument
save_state(Name) :-
   save_state(Name,trale_startup).


% Auxiliary predicates:
save_state(Name,Startup) :-
   current_prolog_flag(compiling,CompileFlag),
   (   set_prolog_flag(compiling,fastcode) % Sparc only
   ->  true                     
   ;   set_prolog_flag(compiling,compactcode)
   ),
   absolute_file_name(Name,AbsName),
   add_state_suffix(AbsName,FullName),
   save_program(FullName,user:Startup),
   set_prolog_flag(compiling,CompileFlag).

%% add_state_suffix/2: append state suffix
%% Adds hosttype to name of state so that we can have different states
%% for each operating system. 
add_state_suffix(In,Out) :-
   (   environ('STATE_SUFFIX',Suffix) 
   ->  atom_concat(In,Suffix,Out)
   ;   In=Out
   ).

%% load_state/1: loading state
load_state(Name):-
   absolute_file_name(Name,AbsName),
   add_state_suffix(AbsName,FullName),
   restore(FullName).

% -------------------------------------------------------------------
% Enhanced query input as frontend to rec/2 
%
% rec/0 prompts user for a string, ending with a newline
%  and passes the appropriate word list to rec/1
%  with a variable Desc argument.

rec :-
   read_line_of_words(List,Desc),
   rec(List,Desc).

% rec_desc/1 work like rec/0, but with a Desc argument - Only parses
% matching the description are returned.

rec_desc(Desc) :-
   read_line_of_words(List,Desc1),
   rec(List,(Desc,Desc1)).

% ------------------------------

read_line_of_words(List,Desc) :-
  read_line(CharList),
  general_tokenize_sentence_string(CharList,List,Desc).

% general_tokenize_sentence_atom(+Atom,-Tokenlist,-StartSymbolDesc)
general_tokenize_sentence_atom(Atom,Tokenlist,StartSymbolDesc) :-
   atom_codes(Atom,String),
   general_tokenize_sentence_string(String,Tokenlist,StartSymbolDesc).

% general_tokenize_sentence_string(String,-Tokenlist,-StartSymbolDesc)
% This predicate allows the grammar to define the tokenization to 
% be used by defining the predicate
%
% tokenize_sentence_string(+String,-Tokenlist,-StartSymbolDesc)
%
% for a particular grammar. In that one can e.g. specify to lowercase 
% everything automatically, what to do with special punctuation,
% how to treat special symbols like apostrophes, etc.

general_tokenize_sentence_string(ItemInput0,ItemInput,Desc) :-
        (  current_predicate(tokenize_sentence_string/3)
	-> tokenize_sentence_string(ItemInput0,ItemInput,Desc)
	;  tokenize_sentence_string_default(ItemInput0,ItemInput,Desc)
	).

tokenize_sentence_string_default(String,Wordlist,bot) :-
   string2wordlist(String,Wordlist).

% string2wordlist(+,-): a very simple tokenizer, i.e. it translates
% a string into words (a list of atoms), taking any number of spaces 
% as delimiting words
%
% e.g.
% "ab"     -> [ab]
% "ab cd"  -> [ab,cd]
% "ab  cd" -> [ab,cd]
%
% i.e.
% [97,98]              -> [ab]
% [97,98,32,99,100]    -> [ab,cd]
% [97,98,32,32,99,100] -> [ab,cd]

string2wordlist([],[]) :- !.
string2wordlist(InString,AtomList) :-
   one_word(InString,WordList,RestString),
   (  WordList == []
   -> AtomList = Atoms
   ;  name(Atom,WordList),
      AtomList = [Atom|Atoms]
   ),
   string2wordlist(RestString,Atoms).

one_word([],[],[]).             % end of string reached
one_word([H|T],Word,Rest) :-
   (  H == 32                        % if space ends a word
   -> Word=[],remove_spaces(T,Rest)  %    remove, incl. additional spaces
   ;  one_char(H,Word0,Word),	     % else map one char
      one_word(T,Word0,Rest)         %      and continue on word
   ).

one_char(X,L,[X|L]).

remove_spaces([],[]).
remove_spaces([H|T],R) :-
   (  H == 32 
   -> remove_spaces(T,R)
   ;  [H|T]=R
   ).

% ------------------------------------------------------------------------

% wordlist2string(+,-)
%
% translate a wordlist into a string. This is almost the inverse
% of string2wordlist/2, but with the difference that always insert
% exactly one space between each word (and never any space at the
% beginning or the end of the string).

wordlist2string([],[]).

wordlist2string([Atom|Atoms],EndPointer0) :-
        name(Atom,WordList),
        makedifflist(WordList,EndPointer0,EndPointer1),
        (Atoms == [] ->
                 EndPointer1 = EndPointer2
        ;   EndPointer1 = [32|EndPointer2]),
        wordlist2string(Atoms,EndPointer2).
        
makedifflist([],X,X).
makedifflist([H|T],[H|L],Diff) :-
        makedifflist(T,L,Diff).


%%% --------------------------------------------------------------- %%
%%% Additional user predicates ==---------------------------------- %%
%%% --------------------------------------------------------------- %%

% Define more user friendly (i.e. less typing) query_proceed/0.
:- abolish(query_proceed/0).

query_proceed :-
   query_user('Do you want to try for more solutions? (return for <~w>) ',[yes],yes,Answer),
   ( ( Answer == y
     ; Answer == 'y.'
     ; Answer == yes
     ; Answer == 'yes.'
     ) -> fail 
   ; true 
   ).

%% query_user(+Format,+Args,+Default,-Value).
%%
%% Useful for getting user input, for instance:
%%
%% query_user('File name? (return for `~w'') : ',[foo],foo,File).
%%
query_user(Format,Args,Default,Value) :-
   nl,
   format(Format,Args),
   get_line(Answer0),
   nl,
   ( Answer0 == [] -> 
                 Value = Default
        ; atom_codes(Value,Answer0)
   ).

%------------------------------------------------------------------------
% predicates for storing and accessing titles that can be displayed 
% by some of the pretty printers (e.g. the one using the grisu interface)

:- dynamic title/1.

clear_title :-
   retractall(title(_)).

set_title(Title) :-
   \+ \+ (
   numbervars(Title,0,_),
   retractall(title(_)),
   assert(title(Title))
         ).
get_title(Title) :-
   (  title(Title)
   -> true
   ;  Title=untitled
   ).

%------------------------------------------------------------------------
% lex_desc(Word:<word>,Desc:<description>)
%------------------------------------------------------------------------
% prints out all words matching the description, querying user in between
%------------------------------------------------------------------------

lex_desc(Word,Desc):-
   \+ \+ ( call_residue((lex(Word,FS),
                         extensionalise(FS),
                         deref(FS,Tag,SVs)),Residue),
           ( secret_noadderrs
           ; secret_adderrs,
             fail
           ),
           add_to(Desc,Tag,SVs),
           ( secret_adderrs
           ; secret_noadderrs,
             fail
           ),
           lex_output(Word,Tag,SVs,Residue),
           query_proceed
  ).

lex_output(Word,Tag,SVs,Residue) :-
   ((current_predicate(portray_lex,portray_lex(_,_,_,_)),
             portray_lex(Word,Tag,SVs,Residue)) -> 
             true
           ; nl, write('WORD: '), write(Word),
             nl, write('ENTRY: '),
             pp_fs_res(Tag,SVs,Residue), nl),
   !.

%------------------------------------------------------------------------
% rec_all/1 and rec_all/2 are identical to rec/1 and rec/2 from ale.pl
% except for the fact that they output all solutions without user interaction.
%-----------------------------------------------------------------------------
%
% rec_all(Words:<word>s)
%-----------------------------------------------------------------------------
% basic predicate to parse Words;  prints out all recognized categories
%-----------------------------------------------------------------------------
rec_all(Words):-  
  nl, write('STRING: '),
  nl, number_display(Words,0),
  ttynl, 
  \+ \+ (rec(Words,Tag,SVs,Residue),
	 ((current_predicate(portray_cat,portray_cat(_,_,_,_,_)),  % see also gen/1 - portray_cat/5
	   portray_cat(Words,bot,Tag,SVs,Residue)) -> true         %  can be called with var 1st arg.
	 ; nl, write('CATEGORY: '),nl, ttyflush,
	   pp_fs_res(Tag,SVs,Residue), nl
	 ),
         fail);
  true.

% ----------------------------------------------------------------------------
% rec_all(Words:<word>s,Desc:<desc>)
% ----------------------------------------------------------------------------
% Like rec_all/1, but solution FSs must satisfy Desc
% ----------------------------------------------------------------------------

rec_all(Words,Desc):-  % must add code to print residues
  nl, write('STRING: '),
  nl, number_display(Words,0),
  ttynl, 
  \+ \+ (rec(Words,Tag,SVs,Desc,Residue),
         ((current_predicate(portray_cat,portray_cat(_,_,_,_,_)),
	   portray_cat(Words,Desc,Tag,SVs,Residue)) -> true
	 ; nl, write('CATEGORY: '),nl, ttyflush,
           pp_fs_res(Tag,SVs,Residue),nl
	 ),
         fail);
  true.

%------------------------------------------------------------------------
% Older stuff:


% Save lexicon in a file
save_lexicon(File):-
        open(File,write,Stream),
        findall(_,(lex(String,FS),
                                  write_term(Stream,lex(String,FS),[quoted(true),ignore_ops(true),portrayed(true),cycles(true),max_depth(0)]),
                                  write(Stream,'.'),
                                  nl(Stream)),_),
        flush_output(Stream),
        close(Stream).

read_lexicon(File):-
        open(File,read,Stream),
        repeat,
        read_term(Stream,Entry,[cycles(true)]),
        ( Entry == end_of_file ->
                 close(Stream),!
        ; assert(Entry),
          fail
        ).

/*
% Save lexicon in a data base
:- use_module(library(bdb)).
save_lexicon:-
        db_open_env('ALE lexicon',EnvRef),
        db_open(lexdb,update,lex(+,-),EnvRef,DBRef),
        db_store(DBRef,_,_),
        db_close_env(EnvRef).
*/

compile_gram_no_lex(File) :-
  abolish_preds,
  reconsult(File),
  compile_gram_no_lex.

compile_gram_no_lex :-
  touch('.alec_throw'),
  absolute_file_name('.alec_throw',AbsFileName),
  assert(ale_compiling(AbsFileName)),
  compile_sig_act,
  compile_fun_act,
  compile_cons_act,
  compile_logic_act,
  compile_when_act,
  compile_expand_act,
  compile_dcs_act,
  compile_grammar_no_lex_act,
  retract(ale_compiling(_)).

compile_grammar_no_lex_act:-
  compile_lex_rules_act,
  % here used to be the lexicon compilation
  % read_lexicon(...)
  compile_rules_act,
% 5/1/96 - Octav -- added call for compilation of generation predicate
  compile_generate_act.
