% -*- prolog -*-
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%   $RCSfile: trale-itsdb.pl,v $
%%  $Revision: 1.12 $
%%      $Date: 2003/10/08 20:41:33 $
%%     Author: Stefan Mueller (Stefan.Mueller@uni-bremen.de)
%%    Purpose: Output of TSDB++ statistics, communication
%%             with capi interface, see itsdb.pl
%%             This code is adapted from Frederik Fouvry's
%%             code for ALE
%%             Arity of edge and empty_cat changed
%%             solutions are explicitly stored in solutions/1
%%             by TRALE.
%%   Language: Prolog
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Log: trale-itsdb.pl,v $
%% Revision 1.12  2003/10/08 20:41:33  stefan
%% moved reverse_count_lex_check to ale_redefinitions since
%% it is relevant for both the chart_display and [incr TSDB()].
%%
%% Revision 1.11  2003/10/05 14:16:37  dm
%% Redid the tokenization of the string and calling of the parser.
%% - Tokenization now allows a grammar to specify the tokenization that
%%   should be used.
%% - The parser is now called in a failure driven loop instead of a
%%   findall, which ensures that no unnecessary memory is consumed and
%%   it avoids the problem with the call to solution/1 in the findall
%%   after the call to rec/5, which due to additional backtracking
%%   led to n! number of solutions being reported instead of n.
%%
%% Revision 1.10  2003/08/13 12:06:31  stefan
%% fixed edge_limit check and exception catch in trale-itsdb and ale_redefinitions check and exception catch in trale-itsdb and ale_redefinition
%%
%% Revision 1.9  2003/08/09 15:11:47  dm
%% Changed rec/6 to rec/5 plus lookup of solution/1 to avoid having to
%% have yet another redefinition in ale_redefinitions.pl
%%
%% Revision 1.8  2003/08/09 14:40:58  dm
%% Moved reverse_count_lex_check here from trale_patches.pl (and removed
%% other temporary definition from ale_redefinitions.pl) and modified it
%% to fail when there are errors - instead of raising an exception.
%%
%% Revision 1.7  2003/08/06 10:09:15  stefan
%% I now catch exceptions (for instance missing words, if this
%% is introduced into the set of exception throwing errors again.
%%
%% Revision 1.6  2003/07/09 20:33:07  stefan
%% adapted to new protocol
%%
%% Revision 1.5  2003/05/19 15:30:46  stefan
%% different call to print_mrs. circumvented length limitation for
%% capi_printf in itsdb.pl. Reordered error output and mrs printing
%% so that MRS errors are recorded by [incr TSDB()].
%%
%% Revision 1.4  2003/05/13 09:10:50  stefan
%% put MRS printing in a different file
%%
%% Revision 1.4  2003/05/13 09:10:48  stefan
%% put MRS printing in a different file
%%
%% Revision 1.3  2003/05/10 13:48:53  stefan
%% *** empty log message ***
%%
%% Revision 1.2  2003/05/09 09:21:02  stefan
%% *** empty log message ***
%%
%% Revision 1.1  2003/05/08 21:50:58  stefan
%% *** empty log message ***
%%
%% Revision 1.3  2003/05/08 19:12:58  stefan
%% MRS-Output works
%%
%% Revision 1.2  2003/05/08 09:23:12  stefan
%% before converting print_mrs from working with descriptions to
%% using fs and the code in lr_utils.pl
%%
%%
%% Revision 1.1  2003/05/03 15:30:26  stefan
%% Initial revision
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(library(system),[environ/2]).
:- use_module(library(charsio),[format_to_chars/3,format_to_chars/4]).
:- use_module(library(lists),[memberchk/2,member/2,nth/3,nth0/3]).
:- use_module(library(ordsets),[ord_union/2,list_to_ord_set/2]).

:- use_module(itsdb,[write_field/2,write_field/4,write_table/2]).

:- ensure_loaded([trale_home('tsdb/retract_lex_desc'),    % retraction of stems
                  trale_home('chart_display/print_mrs'),
                  trale_home('chart_display/ale_redefinitions')
                 ]).

:- version('[incr tsdb()] interface for TRALE').

:- dynamic edge_limit/1.

%----------------------------------------------------------------------
% Predicates that actually collect the information from an application
% These are for ALE/TRALE (so not 3.2!)
% They are currently hard-wired in the module user.

% Special action must be taken for compiled predicates to know their
% numbers.  What do we want: grammar numbers or compiled numbers?
user:create_run(_TSInstanceName,_RunID,_Comment,_Interactivep,Protocol,_Command):-
	( callable(Command) -> call(Command)
   ; true
   ),
        retractall(derivation_protocol(_)),
        assert(derivation_protocol(Protocol)),

        
% Apparently supplied by itsdb
%	date_time(Start),
% Apparently supplied by itsdb
%	write_field('run-id',RunID),		% unique test run identifier
% Apparently supplied by itsdb
%	write_field('r-comment',Comment),		% descriptive narrative
        prolog_version_number(Platform),
        write_field('platform',Platform), % implementation platform (version)
% Apparently supplied by itsdb
% 	write_field('tsdb',''),		% tsdb(1) (version) used
        trale_milca_version(TraleMilcaVersion),
        sformat(Application,'Trale ~w',[TraleMilcaVersion]),
 	write_field('application',Application),	% application (version) used
% 	write_field('environment',''),		% application-specific information
	( current_predicate(grammar_version/1) ->
		 grammar_version(Grammar),
		 write_field('grammar',Grammar) % grammar (version) used
	; true
	),
% Not in ALE/TRALE
%	write_field('avms',''),		% number of avm types in image
	number_of(type(_),NTypes),
 	write_field('sorts',NTypes),		% number of sort types in image
 	number_of((_ macro _),NMacros),
 	write_field('templates',NMacros),		% number of templates in image
 	number_of(lex(_,_),NLex),
 	write_field('lexicon',NLex),		% number of lexical entries
 	number_of(lex_rule(_,_),NLex_rule),
 	write_field('lrules',NLex_rule),		% number of lexical rules
 	number_of(rule(_,_),NRule),
 	write_field('rules',NRule),		% number of (non-lexical) rules
% Apparently supplied by itsdb
%  environ('USER',User),
%  write_field('user',User),		% user who did the test run
% Apparently supplied by itsdb
%  host_name(Host),
%  write_field('host',Host),		% machine used for this run
% Apparently supplied by itsdb
% 	shell('uname -rsv',_,OS),
%  write_field('os',OS),		% operating system (version)
% Apparently supplied by itsdb
%  write_field('start',Start),		% start time of this test run
% only in complete_run! - or supplied by itsdb?
% 	write_field('end',''),		% end time for this test run
% Apparently supplied by itsdb
% 	write_field('items',''),		% number of test items in this run
% Apparently supplied by itsdb
% 	write_field('status',''),		% exit status (PVM only)

         % remove unnecessary lexical items
         % the german grammar uses this to get rid of
         % stems which slow down things when zero inflected
         % material occurs in the string.
         ( current_predicate(retract_before_parsing,retract_before_parsing(_)),
             retract_before_parsing(X),
             retract_lex_desc(X)
         ; true
         ),
	true.

% It might be useful to redefine some of the rec predicates for test
% suite purposes (more accurate statistics collection, and removal of all
% pretty printing).

%process_item(1,'er schlft.',_ParseID,1000,_Exhaustivep,Derivationp,X).
user:process_item(ItemID,ItemInput0,_ParseID,MaxEdges,_Exhaustivep,Derivationp,
						_Interactivep):-
	retractall(edge_limit(_)),
	assert(edge_limit(MaxEdges)),

        reset_all,
        general_tokenize_sentence_atom(ItemInput0,ItemInput,Desc),
        nochart_display,
	trimcore,

        memory_used(X),
        assert(initially_memory_used(X)),

	statistics(garbage_collection,[GCs0,_,TGC0]),
	statistics(walltime,[Wall0,_]),
	statistics(runtime,[Run0,_]),

	compute_all_parses(ItemInput,Desc),

	statistics(runtime,[_,T]),

	findall(SolutionId-T,
		solution(SolutionId), 
		Results,
		[fail-RunLastFail]), % this fourth argument adds fail-RunLastFail to end of Results list

	statistics(runtime,[Run1,RunLastFail]),
	statistics(walltime,[Wall1,_]),
	statistics(garbage_collection,[GCs1,_,TGC1]),
% Apparently supplied by itsdb
%	write_field('parse-id',ParseID),		% unique parse identifier
%	write_field('run-id',''),		% test run for this parse
	write_field('i-id',ItemID),		% item parsed

        % write results has to be done before writing the errors since
        % the MRS output called in write results may produce error messages
        
	length(ItemInput,Length),
	used_edges(Length,ListOfTrees), % includes empty edges
	write_results(Results,ListOfTrees,ItemInput), % only _used_ passive edges
        ( error(_,_) ->
            findall(Err,error(_,Err),ErrorList),
            list_to_atom(ErrorList,32,Error),
            write_field('error',Error) % error string (if applicable |:-)

        ; true
        ),
        % in case of errors due to memory or edge limits there could be readings
        ( error(word,_) ->
            Readings = -1
        ; otherwise ->
            length(Results,Readings0),
            Readings is Readings0 - 1
        ),
        write_field('readings',Readings), % number of readings obtained
        
	Results = [FirstId-First|_],
 	( FirstId \== fail ->
 		 write_field('first',First) % time to find first reading (msec)
 	; write_field('first',-1)
 	),
	TTotal is Run1 - Run0,
	write_field('total',TTotal), % total time for parsing (msec)
	TCPU is Run1 - Run0,
	write_field('tcpu',TCPU),		% total (cpu) processing time (msec)
	TGC is TGC1 - TGC0,
	write_field('tgc',TGC),		% gc time used (msec)
	TReal is Wall1 - Wall0,
	write_field('treal',TReal),	  % overall real time (msec)
	number_of(edge(_,_,_,_,_,_,lexicon),Words), % ! does _not_ include empty edges
	write_field('words',Words),	  % lexical entries retrieved
%	write_field('l-stasks',''),		% successful lexical rule applications
%%	write_field('p-ctasks',''),		% parser contemplated tasks (LKB)
%	write_field('p-ftasks',''),		% parser filtered tasks
%	write_field('p-etasks',''),		% parser executed tasks
%	write_field('p-stasks',''),		% parser succeeding tasks
%%	write_field('aedges',''),		% active items in chart (PAGE)
	number_of(edge(_,_,_,_,_,_,_),Pedges1), %! includes lexicon!
	number_of(empty_cat(_,_,_,_,_,_),Pedges2),
	Pedges is Pedges1 + Pedges2,
	write_field('pedges',Pedges),		% passive items in chart
%%	write_field('raedges',''),		% active items contributing to result
	ord_union(ListOfTrees,UsedEdges),
	length(UsedEdges,RPEdges),
	write_field('rpedges',RPEdges),		% passive items contributing to result
%%	write_field('unifications',''),		% number of (node) unifications
%	write_field('copies',''),		% number of (node) copy operations
%	statistics(program,[Heap,_]),

%        statistics(memory,[Memory,_]),   % total size of the process does not change from item to item
                                         % TSDB shows others in performance statistics so it makes
                                         % more sense to store MUsed there
        initially_memory_used(MIni),
	compute_max_memory,
        max_memory_used(MMax),
        
        MUsed is MMax - MIni,
%	write_field('conses',MUsed), %Heap),		% cons() cells allocated
        write_field('others',MUsed),
%	statistics(atoms,[Symbols,_,_]),
%	write_field('symbols',Symbols),		% symbols allocated


        
%	write_field('others',Memory),		% bytes of memory allocated
	GCs is GCs1 - GCs0,
	write_field('gcs',GCs),		% number of garbage collections
% Apparently supplied by itsdb
%	write_field('i-load',''),		% initial load (start of parse)
% Apparently supplied by itsdb
%	write_field('a-load',''),		% average load
% Apparently supplied by itsdb
%   date_time(Date),
%	write_field('date',Date),		% date and time of parse
	% ALE error predicates should be changed for this

%	write_field('comment',''),		% application-specific comment
	( Derivationp == 1 ->
	    % here _all_ passive edges should be stored
	    true
	; true
	),
	write_rules,
	write_output,
	true.


% ALE uses a chart parser, so we here don't differentiate between first and
% all solutions but just call it in a failure driven loop. rec/5 asserts
% the result edges under predicate solution/1
compute_all_parses(ItemInput,Desc) :-
   catch(rec(ItemInput,_FS,_SVs,Desc,_Residue),
	 Exception,
	 ( sformat(String,'~w',[Exception]),
	   assert(error(exception,String)),
	   fail
	 )),
   fail
   ;
   true.


:- abolish([tokenize_sentence_string_default/3]). % defined in trale_extras.pl
 
tokenize_sentence_string_default(StringList,ItemInput,Desc) :- % Stefan M"uller's version
        convert_s(StringList,WordList),
        list2list_desc(WordList,ItemInput,Desc).
%        write(user_error,WordList),nl(user_error)


% Any switches control writing tables?
write_results(Res,UsedEdges,ItemInput):-
	findall(Rec,(
	nth(ResultId,Res,RId-RTime),
	RId \== fail,
% Apparently supplied by itsdb
%	write_field('parse-id',_ParseID),  % parse for this result
	write_field('result-id',ResultId,Rec,C1), % unique result identifier
	% get from list of results
	write_field('time',RTime,C1,C2),		  % time to find this result (msec)
%	write_field('r-ctasks',''),  % parser contemplated tasks
%	write_field('r-ftasks',''),  % parser filtered tasks
%	write_field('r-etasks',''),  % parser executed tasks
%	write_field('r-stasks',''),  % parser succeeding tasks
	% get somehow the size of the term
	write_field('size',-1,C2,C3),		  % size of feature structure
%	write_field('r-aedges',''),  % active items for this result
   % get the number of passive edges for this result only
	findall(L,(member(Tree,UsedEdges),
				  memberchk(RId,Tree),
				  length(Tree,L)),[RPEdges]),
	write_field('r-pedges',RPEdges,C3,C4),  % passive items in this result
	find_derivation([RId],ItemInput,DerivC,[]),  % Or can this be anything?
        atom_codes(Derivation,DerivC),


%       1 fuer alles wie immer und 2 fuer UDF 2.0.  UDF (Unified Derivation Format) 2.0
                     
	(derivation_protocol(1) ->
            write_field('derivation',Derivation,C4,C5)
        ;   write_field('derivation','',C4,C5)  % C4=C5 % We don't know this format
        ),                       % derivation tree for this reading
%	write_field('tree',''),		  % phrase structure tree (CSLI labels)

        edge(RId,_From,_To,Ref,SVs,_Daughters,_Rule),
        print_mrs(Ref-SVs,MRSC,[]),
        atom_codes(MRS,MRSC),
	write_field('mrs',MRS,C5,Cn),		  % mrs for this reading
	Cn = [],
	true),Results),
	(Results \== [] ->
            write_table('results',Results)
        ; true
        ),
	true.

write_rules:-
	findall(Rec,(
% Not supplied by itsdb?
%	write_field('parse-id',ParseID,Rec,C1), % parse for this rule summary
	Rec = C1,
	rule(Rule,_),
	write_field('rule',Rule,C1,C2), % rule name
% 	write_field('filtered','',C2,C3),  % rule postulations filtered
% 	write_field('executed','',C3,C4),  % rule postulations executed
% 	write_field('successes','',C4,C5), % successful rule postulations
% 	write_field('actives','',C5,C6),	  % active edges built from this rule
	C2 = C6,
	number_of(edge(_,_,_,_,_,_,Rule),Passives),
	write_field('passives',Passives,C6,Cn), % passive edges built from this rule
	Cn = [],
	true),Rules),
%	write_table('rules',Rules),
        write_table('statistics',Rules),
	true.

% When and how?
write_output:-
% Not supplied by itsdb?
%   write_field('i-id',-1),	% item for this output specification
   write_field('o-application',''),	% applicable appliaction(s)
	( current_predicate(grammar_version/1) ->
		 grammar_version(Grammar)
	; Grammar = ''
	),
   write_field('o-grammar',Grammar),	% applicable grammar(s)
   write_field('o-ignore',''),	% ignore this item flag
   write_field('o-wf',-1),	% application-specific grammaticality
   write_field('o-gc',-1),	% maximal number of garbage collections
   write_field('o-derivation',''),	% expected derivation
   write_field('o-tree',''),	% expected phrase structure tree
   write_field('o-mrs',''),	% expected mrs
   write_field('o-edges',-1),	% maximal number of edges to build
   environ('USER',User),
   write_field('o-user',User),	% author of this output specification
	itsdb:date_time(Date),
   write_field('o-date',Date),	% creation date
	true.

% Not sure if these are output or input tables ...
write_tree:-
% Probably supplied by itsdb
%  write_field('parse-id',_),
%	write_field('t-version',_),
%	write_field('t-active',_),
%	write_field('t-confidence',_),
%	write_field('t-author',_),
%	write_field('t-start',_),
%	write_field('t-end',_),
%	write_field('t-comment',''),
	true.

write_decision:-
% Probably supplied by itsdb
%	write_field('parse-id',_),
%	write_field('version',_),
%	write_field('d-state',_),
%	write_field('d-type',_),
%	write_field('d-key',_),
%	write_field('d-value',_),
%	write_field('d-start',_),
%	write_field('d-end',_),
%	write_field('d-date',_),
	true.

user:complete_run(_RunID,_AppSpecific):-
% Apparently supplied by itsdb
% 	date_time(End),
%  	write_field('end',End),		% end time for this test run
	true.

user:reconstruct_item(_):-
	true.

%---------------------------------------------------------------------
% Auxiliary predicates

number_of(Predicate,Number):-
	( current_predicate(_,Predicate) ->
		 findall(_,Predicate,List),
		 length(List,Number)
	; Number is 0
	).

used_edges(Length,UsedEdges):-
	findall(Set,(solution(Top),
                     edge(Top,0,Length,_,_,D,_),
		     recreate_tree(D,Edges,[Top]),
		     list_to_ord_set(Edges,Set)),UsedEdges).
	
recreate_tree([],List,List).
recreate_tree([Edge|Sisters],Edges,Rest0):-
	( integer(Edge) ->
		 edge(Edge,_,_,_,_,Daughters,_)
	; Edge = empty(Nr,Pos),
	  empty_cat(Nr,Pos,_,_,Daughters,_)
	),
	recreate_tree(Daughters,Rest,[Edge|Rest0]),
	recreate_tree(Sisters,Edges,Rest).

% Similar to recreate_tree, only rule names are collected instead of edge numbers
% and the tree shape is retained.
find_derivation([],_,Deriv,Deriv).
find_derivation([Edge|Sisters],String,Deriv,Derivn):-
	( integer(Edge) ->
		 edge(Edge,From,To,_,_,Daughters,Rule),
		 ( Rule == lexicon -> nth0(From,String,Name) ; Rule = Name ),
	    format_to_chars(' (\"~a\" ~d ~d',[Name,From,To],Deriv,Deriv1)
	; Edge = empty(Nr,Pos),
	  empty_cat(Nr,Pos,_,_,Daughters,Rule),
            format_to_chars(' (\"~a\" ~d ~d',[Rule,Pos,Pos],Deriv,Deriv1)
	),
	find_derivation(Daughters,String,Deriv1,Deriv2),
	format_to_chars(')',[],Deriv2,Deriv3),
	find_derivation(Sisters,String,Deriv3,Derivn).

reset_all :-
  clear,
  reset_timer,
  trimcore,
  reset_gc_stat,
  reset_max_memory.


:- use_module(library(system),[popen/3]).

shell(Command,_Status,Atom):-
	system:popen(Command,read,Stream),
	read_stream(Stream,Charlist),
	close(Stream),
	atom_codes(Atom,Charlist).

read_stream(Stream,Charlist):-
	get_code(Stream,C),
	(C == -1 -> Charlist = []
	; Charlist = [C|Charlist0],
	  read_stream(Stream,Charlist0)
	).

normalise_input_string([],[]).
normalise_input_string([W|Ws],[NW|NWs]):-
	atom_codes(W,[C|R]),
	( lower_upper_tab(NC,C) ->
		 atom_codes(NW,[NC|R])
	; NW = W
	),
	normalise_input_string(Ws,NWs).




% ==================================================================
% Local Variables:
% outline-regexp: "^%%%+"
% time-stamp-format: "%04y/%02m/%02d %02H:%02M:%02S %Z %u"
% time-stamp-active: t
% End:
